"""Constructors for initializing FlatSampleReader and NestedSampleReader columns
"""
import warnings
from _weakref import proxy
from collections import defaultdict
from typing import Union

from wrapt import ObjectProxy

from .common import open_file_handles
from .layout_flat import FlatSampleReader, FlatSampleWriter
from .layout_nested import (
    FlatSubsampleReader, FlatSubsampleWriter,
    NestedSampleReader, NestedSampleWriter,
)
from ..records.queries import RecordQuery
from ..records import hash_data_db_key_from_raw_key
from ..typesystem import NdarrayFixedShape, NdarrayVariableShape, StringVariableShape
from ..backends import BACKEND_IS_LOCAL_MAP, backend_decoder


# --------------- methods common to all column layout types -------------------


KeyType = Union[str, int]

_column_definitions = (NdarrayVariableShape, NdarrayFixedShape, StringVariableShape)


def column_type_object_from_schema(schema: dict):
    for c in _column_definitions:
        try:
            instance = c(**schema)
            return instance
        except (TypeError, ValueError):
            pass
    else:  # N.B. for-else loop (ie. "no-break")
        raise ValueError(f'Could not instantiate column schema object for {schema}')


def _warn_remote(aset_name):
    warnings.warn(
        f'Column: {aset_name} contains `reference-only` samples, with '
        f'actual data residing on a remote server. A `fetch-data` '
        f'operation is required to access these samples.', UserWarning)


# --------- FlatSampleReader constructor metaclass / setup methods ------------------


def _flat_load_sample_keys_and_specs(column_name, txnctx):
    """Load flat sample key / backend location mapping info memory.

    Parameters
    ----------
    column_name: str
        name of the column to load.
    txnctx: ColumnTxn
        transaction context object used to access commit ref info on disk

    Returns
    -------
    Tuple[FlatSampleMapType, Set[str]]
        First element is single level dictionary mapping sample key to backend
        location. Second element is set of all unique backends encountered
        for every data pice in the column.
    """
    seen_bes = set()
    sspecs = {}
    with txnctx.read() as ctx:
        hashTxn = ctx.hashTxn
        asetNamesSpec = RecordQuery(ctx.dataenv).column_data_records(column_name)
        for asetNames, dataSpec in asetNamesSpec:
            hashKey = hash_data_db_key_from_raw_key(dataSpec.digest)
            hash_ref = hashTxn.get(hashKey)
            be_loc = backend_decoder(hash_ref)
            sspecs[asetNames.sample] = be_loc
    seen_bes.update((spc.backend for spc in sspecs.values()))
    return (sspecs, seen_bes)


def generate_flat_column(txnctx, column_name, path, schema, mode):
    """Generate instance ready structures for read-only checkouts

    Parameters
    ----------
    txnctx : ColumnTxn
        transaction context object used to access commit ref info on disk
    column_name : str
        name of the column that the reader constructors are being
        generated for
    path : Path
        path to the repository on disk
    schema : ColumnDefinitionTypes
        schema definition of the column.
    mode: str
        read-only or write-enabled mode. one of ['a', 'r'].

    Returns
    -------
    :class:`~.flat.FlatSampleReader`
        Top level column accessor classes fully initialized for requested
        state. initailized structures defining and initializing access to
        the sample data on disk.
    """
    sspecs, bes = _flat_load_sample_keys_and_specs(column_name, txnctx)
    if not all([BACKEND_IS_LOCAL_MAP[be] for be in bes]):
        _warn_remote(column_name)
    if mode == 'a':
        bes.add(schema.backend)
    file_handles = open_file_handles(backends=bes, path=path, mode=mode, schema=schema)

    if mode == 'r':
        res = FlatSampleReader(columnname=column_name,
                               samples=sspecs,
                               backend_handles=file_handles,
                               schema=schema,
                               repo_path=path,
                               mode=mode)
    elif mode == 'a':
        res = FlatSampleWriter(aset_ctx=txnctx,
                               columnname=column_name,
                               samples=sspecs,
                               backend_handles=file_handles,
                               schema=schema,
                               repo_path=path,
                               mode=mode)
    else:
        raise ValueError(f'mode {mode} is not valid.')

    return res


# --------- NestedSampleReader constructor metaclass / setup methods ----------------


def _nested_load_sample_keys_and_specs(column_name, txnctx):
    """Load nested sample/subsample keys and backend location into memory from disk.

    Parameters
    ----------
    column_name : str
        name of the column to load.
    txnctx : ColumnTxn
        transaction context object used to access commit ref info on disk

    Returns
    -------
    Tuple[NestedSampleMapType, Set[str]]
        First element is nested dictionary where each sample name maps to
        subsample contents dict (associating subsample names with backend
        locations). Second element is set of all unique backends encountered
        for every data pice in the column.
    """
    seen_bes = set()
    sspecs = defaultdict(dict)
    with txnctx.read() as ctx:
        hashTxn = ctx.hashTxn
        asetNamesSpec = RecordQuery(ctx.dataenv).column_data_records(column_name)
        for asetNames, dataSpec in asetNamesSpec:
            hashKey = hash_data_db_key_from_raw_key(dataSpec.digest)
            hash_ref = hashTxn.get(hashKey)
            be_loc = backend_decoder(hash_ref)
            sspecs[asetNames.sample].update({asetNames.subsample: be_loc})
            seen_bes.add(be_loc.backend)
    return (sspecs, seen_bes)


def generate_nested_column(txnctx, column_name, path, schema, mode):
    """Generate instance ready structures for read-only checkouts

    Parameters
    ----------
    txnctx : ColumnTxn
        transaction context object used to access commit ref info on disk
    column_name : str
        name of the column that the reader constructors are being
        generated for
    path : Path
        path to the repository on disk
    schema : ColumnDefinitionTypes
        schema definition of the column.
    mode: str
    read-only or write-enabled mode. one of ['a', 'r'].

    Returns
    -------
    :class:`~.nested.NestedSampleReader`
        Top level column accessor classes fully initialized for requested
        state. Initailized structures defining and initializing access to
        the subsample data on disk.
    """
    specs, bes = _nested_load_sample_keys_and_specs(column_name, txnctx)
    if not all([BACKEND_IS_LOCAL_MAP[be] for be in bes]):
        _warn_remote(column_name)
    if mode == 'a':
        bes.add(schema.backend)
    fhand = open_file_handles(backends=bes, path=path, mode=mode, schema=schema)
    samples = {}
    schema_proxy = proxy(schema)
    fhand['enter_count'] = 0

    if mode == 'r':
        for samp, subspecs in specs.items():
            samples[samp] = FlatSubsampleReader(
                columnname=column_name,
                samplen=samp,
                be_handles=fhand,
                specs=subspecs,
                mode='r')
        res = NestedSampleReader(
            columnname=column_name,
            samples=samples,
            backend_handles=fhand,
            repo_path=path,
            mode='r',
            schema=schema)
    elif mode == 'a':
        fhand = ObjectProxy(fhand)
        fhand_proxy = proxy(fhand)
        for samp, subspecs in specs.items():
            samples[samp] = FlatSubsampleWriter(
                schema=schema_proxy,
                aset_ctx=proxy(txnctx),
                repo_path=path,
                columnname=column_name,
                samplen=samp,
                be_handles=fhand_proxy,
                specs=subspecs,
                mode='a')
        res = NestedSampleWriter(
            aset_ctx=txnctx,
            columnname=column_name,
            samples=samples,
            backend_handles=fhand,
            schema=schema,
            repo_path=path,
            mode='a')
    else:
        raise ValueError(f'mode {mode} is not valid.')

    return res



"""Module that contains the command line app.

Why does this file exist, and why not put this in __main__?

   You might be tempted to import things from __main__ later, but that will cause
   problems: the code will get executed twice:

      - When you run `python -mhangar` python will execute
        ``__main__.py`` as a script. That means there won't be any
        ``hangar.__main__`` in ``sys.modules``.
      - When you import __main__ it will get executed again (as a module) because
        there's no ``hangar.__main__`` in ``sys.modules``.

Also see (1) from http://click.pocoo.org/7/setuptools/#setuptools-integration
"""
import os
import time
from pathlib import Path

import click
import numpy as np

from hangar import Repository, __version__

from .utils import parse_custom_arguments, StrOrIntType


pass_repo = click.make_pass_decorator(Repository, ensure=True)


@click.group(no_args_is_help=True, add_help_option=True, invoke_without_command=True)
@click.version_option(version=__version__, help='display current Hangar Version')
@click.pass_context
def main(ctx):  # pragma: no cover
    P = os.getcwd()
    ctx.obj = Repository(path=P, exists=False)


# -------------------------------- Init ---------------------------------------


@main.command()
@click.option('--name', prompt='User Name', help='first and last name of user')
@click.option('--email', prompt='User Email', help='email address of the user')
@click.option('--overwrite', is_flag=True, default=False,
              help='overwrite a repository if it exists at the current path')
@pass_repo
def init(repo: Repository, name, email, overwrite):
    """Initialize an empty repository at the current path
    """
    if repo.initialized and (not overwrite):
        click.echo(f'Repo already exists at: {repo.path}')
    else:
        repo.init(user_name=name, user_email=email, remove_old=overwrite)


# -------------------------- Writer Lock -------------------------------------


@main.command(name='writer-lock')
@click.option('--force-release', 'force_release_', is_flag=True, default=False,
              help='force release writer lock from the CLI.')
@pass_repo
def writer_lock_held(repo: Repository, force_release_):
    """Determine if the writer lock is held for a repository.

    Passing the --force-release flag will instantly release the writer lock,
    invalidating any process which currently holds it.
    """
    if force_release_:
        repo.force_release_writer_lock()
        click.echo(f'Success force release of writer lock.')
    else:
        if repo.writer_lock_held:
            click.echo(f'Writer lock is held.')
        else:
            click.echo(f'Writer lock is available.')



# -------------------------- Checkout Writer ----------------------------------


@main.command()
@click.argument('branchname', nargs=1, required=True)
@pass_repo
def checkout(repo: Repository, branchname):
    """Checkout writer head branch at BRANCHNAME.

    This method requires that no process currently holds the writer lock.
    In addition, it requires that the contents of the staging area are
    'CLEAN' (no changes have been staged).
    """
    try:
        co = repo.checkout(write=True, branch=branchname)
        co.close()
        click.echo(f'Writer checkout head set to branch: {branchname}')
    except (ValueError, PermissionError) as e:
        raise click.ClickException(e)


@main.command()
@click.option('--message', '-m', multiple=True,
              help=('The commit message. If provided multiple times '
                    'each argument gets converted into a new line.'))
@pass_repo
def commit(repo: Repository, message):
    """Commits outstanding changes.

    Commit changes to the given files into the repository. You will need to
    'push' to push up your changes to other repositories.
    """
    from hangar.records.summarize import status

    co = repo.checkout(write=True)
    try:
        if not message:
            diff = co.diff.staged()
            status_txt = status(co._hashenv, co.branch_name, diff.diff)
            status_txt.seek(0)
            marker = '# Changes To Be committed: \n'
            hint = ['\n', '\n', marker, '# \n']
            for line in status_txt.readlines():
                hint.append(f'# {line}')
            # open default system editor
            message = click.edit(''.join(hint))
            if message is None:
                click.echo('Aborted!')
                return
            msg = message.split(marker)[0].rstrip()
            if not msg:
                click.echo('Aborted! Empty commit message')
                return
        else:
            msg = '\n'.join(message)

        click.echo('Commit message:\n' + msg)
        try:
            digest = co.commit(msg)
            click.echo(f'Commit Successful. Digest: {digest}')
        except RuntimeError as e:
            raise click.ClickException(e)
    finally:
        co.close()


# -------------------------- Column Interactor ------------------------------


@main.group(no_args_is_help=True, add_help_option=True)
@click.pass_context
def column(ctx):  # pragma: no cover
    """Operations for working with columns in the writer checkout.
    """
    pass


@column.command(name='create')
@click.option('--variable-shape', 'variable_', is_flag=True, default=False,
              help='flag indicating sample dimensions can be any size up to max shape.')
@click.option('--contains-subsamples', 'subsamples_', is_flag=True, default=False,
              help=('flag indicating if this is a column which nests multiple '
                    'subsamples under a common sample key.'))
@click.argument('name', nargs=1, type=click.STRING, required=True)
@click.argument('dtype', nargs=1, type=click.Choice([
    'UINT8', 'INT8', 'UINT16', 'INT16', 'UINT32', 'INT32',
    'UINT64', 'INT64', 'FLOAT16', 'FLOAT32', 'FLOAT64', 'STR']), required=True)
@click.argument('shape', nargs=-1, type=click.INT, required=False)
@pass_repo
def create_column(repo: Repository, name, dtype, shape, variable_, subsamples_):
    """Create an column with NAME and DTYPE of SHAPE.

    The column will be created in the staging area / branch last used by a
    writer-checkout. Valid NAMEs contain only ascii letters and [``'.'``,
    ``'_'``, ``'-'``] (no whitespace). The DTYPE must be one of [``'UINT8'``,
    ``'INT8'``, ``'UINT16'``, ``'INT16'``, ``'UINT32'``, ``'INT32'``,
    ``'UINT64'``, ``'INT64'``, ``'FLOAT16'``, ``'FLOAT32'``, ``'FLOAT64'``,
    ``'STR'``].

    If a ndarray dtype is specified (not 'STR'), then the SHAPE must be the
    last argument(s) specified, where each dimension size is identified by
    a (space seperated) list of numbers.

    Examples:

    To specify, an column for some training images of dtype uint8 and shape
    (256, 256, 3) we should say:

       .. code-block:: console

          $ hangar column create train_images UINT8 256 256 3

    To specify that the samples can be variably shaped (have any dimension size
    up to the maximum SHAPE specified) we would say:

       .. code-block:: console

          $ hangar column create train_images UINT8 256 256 3 --variable-shape

    or equivalently:

       .. code-block:: console

          $ hangar column create --variable-shape train_images UINT8 256 256 3

    To specify that the column contains a nested set of subsample data under a
    common sample key, the ``--contains-subsamples`` flag can be used.

       .. code-block:: console

          $ hangar column create --contains-subsamples train_images UINT8 256 256 3

    """
    try:
        co = repo.checkout(write=True)
        if dtype == 'STR':
            col = co.add_str_column(name=name, contains_subsamples=subsamples_)
        else:
            col = co.add_ndarray_column(name=name,
                                        shape=shape,
                                        dtype=np.typeDict[dtype.lower()],
                                        variable_shape=variable_,
                                        contains_subsamples=subsamples_)
        click.echo(f'Initialized Column: {col.column}')
    except (ValueError, LookupError, PermissionError) as e:
        raise click.ClickException(e)
    finally:
        try:
            co.close()
        except NameError:
            pass


@column.command(name='remove')
@click.argument('name', nargs=1, type=click.STRING, required=True)
@pass_repo
def remove_column(repo: Repository, name):
    """Delete the column NAME (and all samples) from staging area.

    The column will be removed from the staging area / branch last used by a
    writer-checkout.
    """
    try:
        co = repo.checkout(write=True)
        removed = co.columns.delete(name)
        click.echo(f'Successfully removed column: {removed}')
    except (ValueError, KeyError, PermissionError) as e:
        raise click.ClickException(e)
    finally:
        try:
            co.close()
        except NameError:
            pass


# ---------------------------- Remote Interaction -----------------------------


@main.command()
@click.argument('remote', nargs=1, required=True)
@click.option('--name', prompt='User Name', help='first and last name of user')
@click.option('--email', prompt='User Email', help='email address of the user')
@click.option('--overwrite', is_flag=True, default=False,
              help='overwrite a repository if it exists at the current path')
@pass_repo
def clone(repo: Repository, remote, name, email, overwrite):
    """Initialize a repository at the current path and fetch updated records from REMOTE.

    Note: This method does not actually download the data to disk. Please look
    into the ``fetch-data`` command.
    """
    if repo.initialized and (not overwrite):
        click.echo(f'Repo already exists at: {repo.path}')
    else:
        repo.clone(name, email, remote, remove_old=overwrite)


@main.command(name='fetch')
@click.argument('remote', nargs=1, required=True)
@click.argument('branch', nargs=1, required=True)
@pass_repo
def fetch_records(repo: Repository, remote, branch):
    """Retrieve the commit history from REMOTE for BRANCH.

    This method does not fetch the data associated with the commits. See
    ``fetch-data`` to download the tensor data corresponding to a commit.
    """
    bName = repo.remote.fetch(remote=remote, branch=branch)
    click.echo(f'Fetched branch Name: {bName}')


@main.command(name='fetch-data')
@click.argument('remote', nargs=1, required=True)
@click.argument('startpoint', nargs=1, required=True)
@click.option('--column', '-d', multiple=True, required=False, default=None,
              help='specify any number of column keys to fetch data for.')
@click.option('--nbytes', '-n', default=None, required=False,
              help='total amount of data to retrieve in MB/GB.')
@click.option('--all-history', '-a', 'all_', is_flag=True, default=False, required=False,
              help='Retrieve data referenced in every parent commit accessible to the STARTPOINT')
@pass_repo
def fetch_data(repo: Repository, remote, startpoint, column, nbytes, all_):
    """Get data from REMOTE referenced by STARTPOINT (short-commit or branch).

    The default behavior is to only download a single commit's data or the HEAD
    commit of a branch. Please review optional arguments for other behaviors
    """
    from hangar.records.commiting import expand_short_commit_digest
    from hangar.records.heads import get_branch_head_commit
    from hangar.records.heads import get_staging_branch_head
    from hangar.utils import parse_bytes

    if startpoint is None:
        branch = get_staging_branch_head(repo._env.branchenv)
        commit = get_branch_head_commit(repo._env.branchenv, branch)
    elif startpoint in repo.list_branches():
        commit = get_branch_head_commit(repo._env.branchenv, startpoint)
    else:
        commit = expand_short_commit_digest(repo._env.refenv, startpoint)
    click.echo(f'Fetching data for commit: {commit}')

    try:
        max_nbytes = parse_bytes(nbytes)
    except AttributeError:
        max_nbytes = None
    if len(column) == 0:
        column = None

    commits = repo.remote.fetch_data(remote=remote,
                                     commit=commit,
                                     column_names=column,
                                     max_num_bytes=max_nbytes,
                                     retrieve_all_history=all_)
    click.echo(f'completed data for commits: {commits}')


@main.command()
@click.argument('remote', nargs=1, required=True)
@click.argument('branch', nargs=1, required=True)
@pass_repo
def push(repo: Repository, remote, branch):
    """Upload local BRANCH commit history / data to REMOTE server.
    """
    commit_hash = repo.remote.push(remote=remote, branch=branch)
    click.echo(f'Push data for commit hash: {commit_hash}')


# ----------------------- Remote Server References ----------------------------


@main.group(no_args_is_help=True, add_help_option=True)
@click.pass_context
def remote(ctx):  # pragma: no cover
    """Operations for working with remote server references
    """
    pass


@remote.command(name='list')
@pass_repo
def list_remotes(repo: Repository):
    """List all remote repository records.
    """
    click.echo(repo.remote.list_all())


@remote.command(name='add')
@click.argument('name', nargs=1, required=True)
@click.argument('address', nargs=1, required=True)
@pass_repo
def add_remote(repo: Repository, name, address):
    """Add a new remote server NAME with url ADDRESS to the local client.

    This name must be unique. In order to update an old remote, please remove it
    and re-add the remote NAME / ADDRESS combination
    """
    click.echo(repo.remote.add(name=name, address=address))


@remote.command(name='remove')
@click.argument('name', nargs=1, required=True)
@pass_repo
def remove_remote(repo: Repository, name):
    """Remove the remote server NAME from the local client.

    This will not remove any tracked remote reference branches.
    """
    click.echo(repo.remote.remove(name=name))


# ---------------------------- User Visualizations ----------------------------


@main.command()
@click.argument('dev', nargs=1, required=True)
@click.argument('master', nargs=1, required=False, default=None)
@pass_repo
def diff(repo: Repository, dev, master):
    """Display diff of DEV commit/branch to MASTER commit/branch.

    If no MASTER is specified, then the staging area branch HEAD will
    will be used as the commit digest for MASTER. This operation will
    return a diff which could be interpreted as if you were merging
    the changes in DEV into MASTER.

    TODO: VERIFY ORDER OF OUTPUT IS CORRECT.
    """
    from hangar.records.commiting import expand_short_commit_digest
    from hangar.records.commiting import get_staging_branch_head
    from hangar.records.summarize import status

    if dev not in repo.list_branches():
        dev = expand_short_commit_digest(repo._env.refenv, dev)

    if master is None:
        master = get_staging_branch_head(repo._env.branchenv)
    elif master not in repo.list_branches():
        master = expand_short_commit_digest(repo._env.refenv, master)

    diff_spec = repo.diff(master, dev)
    buf = status(hashenv=repo._env.hashenv, branch_name=dev, diff=diff_spec.diff)
    click.echo(buf.getvalue())

@main.command()
@click.argument('startpoint', nargs=1, required=False)
@pass_repo
def summary(repo: Repository, startpoint):
    """Display content summary at STARTPOINT (short-digest or branch).

    If no argument is passed in, the staging area branch HEAD wil be used as the
    starting point. In order to recieve a machine readable, and more complete
    version of this information, please see the ``Repository.summary()`` method
    of the API.
    """
    from hangar.records.commiting import expand_short_commit_digest

    if startpoint is None:
        click.echo(repo.summary())
    elif startpoint in repo.list_branches():
        click.echo(repo.summary(branch=startpoint))
    else:
        base_commit = expand_short_commit_digest(repo._env.refenv, startpoint)
        click.echo(repo.summary(commit=base_commit))


@main.command()
@click.argument('startpoint', required=False, default=None)
@pass_repo
def log(repo: Repository, startpoint):
    """Display commit graph starting at STARTPOINT (short-digest or name)

    If no argument is passed in, the staging area branch HEAD will be used as the
    starting point.
    """
    from hangar.records.commiting import expand_short_commit_digest

    if startpoint is None:
        click.echo(repo.log())
    elif startpoint in repo.list_branches():
        click.echo(repo.log(branch=startpoint))
    else:
        base_commit = expand_short_commit_digest(repo._env.refenv, startpoint)
        click.echo(repo.log(commit=base_commit))


@main.command()
@pass_repo
def status(repo: Repository):
    """Display changes made in the staging area compared to it's base commit
    """
    from hangar.records.summarize import status
    co = repo.checkout(write=True)
    try:
        diff = co.diff.staged()
        click.echo(status(co._hashenv, co.branch_name, diff.diff).getvalue(), nl=False)
    finally:
        co.close()


# ------------------------------- Branching -----------------------------------


@main.group(no_args_is_help=True, add_help_option=True)
@click.pass_context
def branch(ctx):  # pragma: no cover
    """operate on and list branch pointers.
    """
    pass


@branch.command(name='list')
@pass_repo
def branch_list(repo: Repository):
    """list all branch names

    Includes both remote branches as well as local branches.
    """
    click.echo(repo.list_branches())


@branch.command(name='create')
@click.argument('name', nargs=1, required=True)
@click.argument('startpoint', nargs=1, default=None, required=False)
@pass_repo
def branch_create(repo: Repository, name, startpoint):
    """Create a branch with NAME at STARTPOINT (short-digest or branch)

    If no STARTPOINT is provided, the new branch is positioned at the HEAD of
    the staging area branch, automatically.
    """
    from hangar.records.commiting import expand_short_commit_digest
    from hangar.records.heads import get_branch_head_commit
    from hangar.records.heads import get_staging_branch_head

    branch_names = repo.list_branches()
    if name in branch_names:
        e = ValueError(f'branch name: {name} already exists')
        raise click.ClickException(e)

    try:
        if startpoint is None:
            branch = get_staging_branch_head(repo._env.branchenv)
            base_commit = get_branch_head_commit(repo._env.branchenv, branch)
        elif startpoint in branch_names:
            base_commit = get_branch_head_commit(repo._env.branchenv, startpoint)
        else:
            base_commit = expand_short_commit_digest(repo._env.refenv, startpoint)

        res = repo.create_branch(name, base_commit=base_commit)
    except (KeyError, ValueError, RuntimeError) as e:
        raise click.ClickException(e)

    click.echo(f'Created BRANCH: {res.name} HEAD: {res.digest}')


@branch.command(name='delete')
@click.argument('name', nargs=1, required=True)
@click.option('--force', '-f', is_flag=True, default=False,
              help='flag to force delete branch which has un-merged history.')
@pass_repo
def branch_remove(repo: Repository, name, force):
    """Remove a branch pointer with the provided NAME

    The NAME must be a branch present on the local machine.
    """
    try:
        res = repo.remove_branch(name, force_delete=force)
    except (ValueError, PermissionError, RuntimeError) as e:
        raise click.ClickException(e)

    click.echo(f'Deleted BRANCH: {res.name} HEAD: {res.digest}')


# ---------------------------- Server Commands --------------------------------


@main.command()
@click.option('--overwrite', is_flag=True, default=False,
              help='overwrite the hangar server instance if it exists at the current path.')
@click.option('--ip', default='localhost', show_default=True,
              help='the ip to start the server on. default is `localhost`')
@click.option('--port', default='50051', show_default=True,
              help='port to start the server on. default in `50051`')
@click.option('--timeout', default=60 * 60 * 24, required=False, show_default=True,
              help='time (in seconds) before server is stopped automatically')
def server(overwrite, ip, port, timeout):
    """Start a hangar server, initializing one if does not exist.

    The server is configured to top working in 24 Hours from the time it was
    initially started. To modify this value, please see the ``--timeout``
    parameter.

    The hangar server directory layout, contents, and access conventions are
    similar, though significantly different enough to the regular user "client"
    implementation that it is not possible to fully access all information via
    regular API methods. These changes occur as a result of the uniformity of
    operations promised by both the RPC structure and negotiations between the
    client/server upon connection.

    More simply put, we know more, so we can optimize access more; similar, but
    not identical.
    """
    from hangar.remote.server import serve

    P = os.getcwd()
    ip_port = f'{ip}:{port}'
    server, hangserver, channel_address = serve(P, overwrite, channel_address=ip_port)
    server.start()
    click.echo(f'Hangar Server Started')
    click.echo(f'* Start Time: {time.asctime()}')
    click.echo(f'* Base Directory Path: {P}')
    click.echo(f'* Operating on `IP_ADDRESS:PORT`: {channel_address}')
    try:
        startTime = time.time()
        while True:
            time.sleep(0.1)
            if time.time() - startTime > timeout:
                raise SystemExit
    except (KeyboardInterrupt, SystemExit):
        click.echo(f'Server Stopped at Time: {time.asctime()}')
        hangserver.close()
        server.stop(0)


# ---------------------------- Import Exporters -------------------------------


@main.command(name='import',
              context_settings=dict(allow_extra_args=True, ignore_unknown_options=True, ))
@click.argument('column', required=True)
@click.argument('path',
                required=True,
                type=click.Path(exists=True, dir_okay=True, file_okay=True, readable=True,
                                resolve_path=True))
@click.option('--branch', default=None, help='branch to import data')
@click.option('--plugin', default=None, help='override auto-infered plugin')
@click.option('--overwrite', is_flag=True,
              help='overwrite data samples with the same name as the imported data file ')
@pass_repo
@click.pass_context
def import_data(ctx, repo: Repository, column, path, branch, plugin, overwrite):
    """Import file or directory of files at PATH to COLUMN in the staging area.

    If passing in a directory, all files in the directory will be imported, if
    passing in a file, just that files specified will be
    imported
    """
    # TODO: ignore warning through env variable
    from types import GeneratorType
    from hangar import external
    from hangar.records.heads import get_staging_branch_head

    kwargs = parse_custom_arguments(ctx.args)
    if branch is None:
        branch = get_staging_branch_head(repo._env.branchenv)
    elif branch not in repo.list_branches():
        raise click.ClickException(f'Branch name: {branch} does not exist, Exiting.')
    click.echo(f'Writing to branch: {branch}')

    co = repo.checkout(write=True, branch=branch)
    try:
        active_aset = co.columns.get(column)
        p = Path(path)
        files = [f.resolve() for f in p.iterdir()] if p.is_dir() else [p.resolve()]
        with active_aset as aset, click.progressbar(files) as filesBar:
            for f in filesBar:
                ext = ''.join(f.suffixes).strip('.')  # multi-suffix files (tar.bz2)
                loaded = external.load(f, plugin=plugin, extension=ext, **kwargs)
                if not isinstance(loaded, GeneratorType):
                    loaded = [loaded]
                for arr, fname in loaded:
                    if (not overwrite) and (fname in aset):
                        continue
                    try:
                        aset[fname] = arr
                    except ValueError as e:
                        click.echo(e)
    except (ValueError, KeyError) as e:
        raise click.ClickException(e)
    finally:
        co.close()


@main.command(name='export',
              context_settings=dict(allow_extra_args=True, ignore_unknown_options=True, ))
@click.argument('column', nargs=1, required=True)
@click.argument('startpoint', nargs=1, default=None, required=False)
@click.option('-o', '--out', 'outdir',
              nargs=1,
              required=False,
              default=os.getcwd(),
              type=click.Path(exists=True, dir_okay=True, file_okay=False, readable=True,
                              resolve_path=True),
              help="Directory to export data")
@click.option('-s', '--sample',
              nargs=1,
              default=None,
              type=StrOrIntType(),
              help=('Sample name to export. Default implementation is to interpret all input '
                    'names as string type. As an column can contain samples with both ``str`` '
                    'and ``int`` types, we allow you to specify ``name type`` of the sample. To '
                    'identify a potentially ambiguous name, we allow you to prepend the type of '
                    'sample name followed by a colon and then the sample name (ex. ``str:54`` '
                    'or ``int:54``). this can be done for any sample key.'))
@click.option('-f', '--format', 'format_',
              nargs=1,
              required=False,
              help='File format of output file')
@click.option('--plugin', required=False, help='override auto-inferred plugin')
@pass_repo
@click.pass_context
def export_data(ctx, repo: Repository, column, outdir, startpoint, sample, format_, plugin):
    """Export COLUMN sample data as it existed a STARTPOINT to some format and path.

    Specifying which sample to be exported is possible by using the switch
    ``--sample`` (without this, all the samples in the given column will be
    exported). Since hangar supports both int and str datatype for the sample
    name, specifying that while mentioning the sample name might be necessary
    at times. It is possible to do that by separating the name and type by a
    colon.

    Example:

       1. if the sample name is string of numeric 10 - ``str:10`` or ``10``

       2. if the sample name is ``sample1`` - ``str:sample1`` or ``sample1``

       3. if the sample name is an int, let say 10 - ``int:10``
    """
    from hangar.records.commiting import expand_short_commit_digest
    from hangar.records.heads import get_branch_head_commit, get_staging_branch_head
    from hangar import external
    kwargs = parse_custom_arguments(ctx.args)

    if startpoint in repo.list_branches():
        base_commit = get_branch_head_commit(repo._env.branchenv, startpoint)
    elif startpoint:
        base_commit = expand_short_commit_digest(repo._env.refenv, startpoint)
    else:
        branch_name = get_staging_branch_head(repo._env.branchenv)
        base_commit = get_branch_head_commit(repo._env.branchenv, branch_name)

    co = repo.checkout(commit=base_commit)
    try:
        aset = co.columns.get(column)
        sampleNames = [sample] if sample is not None else list(aset.keys())
        extension = format_.lstrip('.') if format_ else None
        with aset, click.progressbar(sampleNames) as sNamesBar:
            for sampleN in sNamesBar:
                data = aset[sampleN]
                formated_sampleN = f'{type(sampleN).__name__}:{sampleN}'
                try:
                    external.save(data, outdir, formated_sampleN, extension, plugin, **kwargs)
                except Exception as e:
                    raise click.ClickException(e)
    except KeyError as e:
        raise click.ClickException(e)
    finally:
        co.close()


@main.command(name='view',
              context_settings=dict(allow_extra_args=True, ignore_unknown_options=True, ))
@click.argument('column', nargs=1, type=str, required=True)
@click.argument('sample', nargs=1, type=StrOrIntType(), required=True)
@click.argument('startpoint', nargs=1, default=None, required=False)
@click.option('-f', '--format', 'format_', required=False, help='File format of output file')
@click.option('--plugin', default=None, help='Plugin name to use instead of auto-inferred plugin')
@pass_repo
@click.pass_context
def view_data(ctx, repo: Repository, column, sample, startpoint, format_, plugin):
    """Use a plugin to view the data of some SAMPLE in COLUMN at STARTPOINT.
    """
    from hangar.records.commiting import expand_short_commit_digest
    from hangar.records.heads import get_branch_head_commit, get_staging_branch_head
    from hangar import external

    kwargs = parse_custom_arguments(ctx.args)
    if startpoint in repo.list_branches():
        base_commit = get_branch_head_commit(repo._env.branchenv, startpoint)
    elif startpoint:
        base_commit = expand_short_commit_digest(repo._env.refenv, startpoint)
    else:
        branch_name = get_staging_branch_head(repo._env.branchenv)
        base_commit = get_branch_head_commit(repo._env.branchenv, branch_name)

    co = repo.checkout(commit=base_commit)
    try:
        aset = co.columns.get(column)
        extension = format_.lstrip('.') if format_ else None
        data = aset[sample]
        try:
            external.show(data, plugin=plugin, extension=extension, **kwargs)
        except Exception as e:
            raise click.ClickException(e)
    except KeyError as e:
        raise click.ClickException(e)
    finally:
        co.close()


# ---------------------------- Developer Utils --------------------------------


@main.command(name='db-view', hidden=True)
@click.option('-a', is_flag=True, help='display all dbs in the repository')
@click.option('-b', is_flag=True, help='display the branch/heads db')
@click.option('-r', is_flag=True, help='display the references db')
@click.option('-d', is_flag=True, help='display the data hash db')
@click.option('-s', is_flag=True, help='display the stage record db')
@click.option('-z', is_flag=True, help='display the staged hash record db')
@click.option('--limit', default=30, help='limit the amount of records displayed before truncation')
@pass_repo
def lmdb_record_details(repo: Repository, a, b, r, d, s, z, limit):
    """DEVELOPER TOOL ONLY

    display key/value pairs making up the dbs
    """
    from hangar.context import Environments
    from hangar.records.summarize import details
    from hangar import constants as c

    if repo._repo_path.is_dir():
        repo_path = repo._repo_path
    elif repo._repo_path.parent.joinpath(c.DIR_HANGAR_SERVER).is_dir():
        repo_path = repo._repo_path.parent.joinpath(c.DIR_HANGAR_SERVER)
    else:
        click.echo(f'NO HANGAR INSTALLATION AT PATH: {repo._repo_path.parent}')
        return

    envs = Environments(pth=repo_path)
    try:
        if a:
            b, r, d, s, z = True, True, True, True, True
        if b:
            click.echo(details(envs.branchenv, line_limit=limit).getvalue())
        if r:
            click.echo(details(envs.refenv, line_limit=limit).getvalue())
        if d:
            click.echo(details(envs.hashenv, line_limit=limit).getvalue())
        if s:
            click.echo(details(envs.stageenv, line_limit=limit).getvalue())
        if z:
            click.echo(details(envs.stagehashenv, line_limit=limit).getvalue())
    finally:
        envs._close_environments()



from functools import reduce
from operator import getitem as op_getitem
from contextlib import ExitStack

# noinspection PyUnresolvedReferences
class GetMixin:
    """Mixin methods for the checkout object classes.

    Used since the read and write enabled checkouts have the same :meth:`__get__`
    and :meth:`get` methods
    """

    def __getitem__(self, index):
        """Dictionary style access to columns and samples

        Checkout object can be thought of as a "dataset" ("dset") mapping a
        view of samples across columns.

            >>> dset = repo.checkout(branch='master')
            >>>
            # Get an column contained in the checkout.
            >>> dset['foo']
            ColumnDataReader
            >>>
            # Get a specific sample from ``'foo'`` (returns a single array)
            >>> dset['foo', '1']
            np.array([1])
            >>>
            # Get multiple samples from ``'foo'`` (returns a list of arrays, in order
            # of input keys)
            >>> dset[['foo', '1'], ['foo', '2'],  ['foo', '324']]
            [np.array([1]), np.ndarray([2]), np.ndarray([324])]
            >>>
            # Get sample from multiple columns, column/data returned is ordered
            # in same manner as input of func.
            >>> dset[['foo', '1'], ['bar', '1'],  ['baz', '1']]
            [np.array([1]), np.ndarray([1, 1]), np.ndarray([1, 1, 1])]
            >>>
            # Get multiple samples from multiple columns\
            >>> keys = [(col, str(samp)) for samp in range(2) for col in ['foo', 'bar']]
            >>> keys
            [('foo', '0'), ('bar', '0'), ('foo', '1'), ('bar', '1')]
            >>> dset[keys]
            [np.array([1]), np.array([1, 1]), np.array([2]), np.array([2, 2])]

        Arbitrary column layouts are supported by simply adding additional members
        to the keys for each piece of data. For example, getting data from a column
        with a nested layout:

            >>> dset['nested_col', 'sample_1', 'subsample_0']
            np.array([1, 0])
            >>>
            # a sample accessor object can be retrieved at will...
            >>> dset['nested_col', 'sample_1']
            <class 'FlatSubsampleReader'>(column_name='nested_col', sample_name='sample_1')
            >>>
            # to get all subsamples in a nested sample use the Ellipsis operator
            >>> dset['nested_col', 'sample_1', ...]
            {'subsample_0': np.array([1, 0]),
             'subsample_1': np.array([1, 1]),
             ...
             'subsample_n': np.array([1, 255])}

        Retrieval of data from different column types can be mixed and combined
        as desired. For example, retrieving data from both flat and nested columns
        simultaneously:

            >>> dset[('nested_col', 'sample_1', '0'), ('foo', '0')]
            [np.array([1, 0]), np.array([0])]
            >>> dset[('nested_col', 'sample_1', ...), ('foo', '0')]
            [{'subsample_0': np.array([1, 0]), 'subsample_1': np.array([1, 1])},
             np.array([0])]
            >>> dset[('foo', '0'), ('nested_col', 'sample_1')]
            [np.array([0]),
             <class 'FlatSubsampleReader'>(column_name='nested_col', sample_name='sample_1')]

        If a column or data key does not exist, then this method will raise a KeyError.
        As an alternative, missing keys can be gracefully handeled by calling :meth:`get()`
        instead. This method does not (by default) raise an error if a key is missing.
        Instead, a (configurable) default value is simply inserted in it's place.

            >>> dset['foo', 'DOES_NOT_EXIST']
            -------------------------------------------------------------------
            KeyError                           Traceback (most recent call last)
            <ipython-input-40-731e6ea62fb8> in <module>
            ----> 1 res = co['foo', 'DOES_NOT_EXIST']
            KeyError: 'DOES_NOT_EXIST'

        Parameters
        ----------
        index
            column name, sample key(s) or sequence of list/tuple of column name,
            sample keys(s) which should be retrieved in the operation.

            Please see detailed explanation above for full explanation of accepted
            argument format / result types.

        Returns
        -------
        :class:`~.columns.column.Columns`
            single column parameter, no samples specified
        Any
            Single column specified, single sample key specified
        List[Any]
            arbitrary columns, multiple samples array data for each sample is
            returned in same order sample keys are received.
        """
        # not using kwargs since this could be in a tight loop.
        # kwargs: default-None, except_missing=True
        return self._get_in(index, None, True)

    def get(self, keys, default=None, except_missing=False):
        """View of sample data across columns gracefully handling missing sample keys.

        Please see :meth:`__getitem__()` for full description. This method is
        identical with a single exception: if a sample key is not present in an
        column, this method will plane a null ``None`` value in it's return
        slot rather than throwing a ``KeyError`` like the dict style access
        does.

        Parameters
        ----------
        keys
            sequence of column name (and optionally) sample key(s) or sequence of
            list/tuple of column name, sample keys(s) which should be retrieved in
            the operation.

            Please see detailed explanation in :meth:`__getitem__()` for full
            explanation of accepted argument format / result types.

        default: Any, optional
            default value to insert in results for the case where some column
            name / sample key is not found, and the `except_missing` parameter
            is set to False.

        except_missing: bool, optional
            If False, will not throw exceptions on missing sample key value.
            Will raise KeyError if True and missing key found.

        Returns
        -------
        :class:`~.columns.column.Columns`
            single column parameter, no samples specified
        Any
            Single column specified, single sample key specified
        List[Any]
            arbitrary columns, multiple samples array data for each sample is
            returned in same order sample keys are received.
        """
        return self._get_in(keys, default, except_missing)

    def _get_in(self, keys, default=None, except_missing=False,
                *, _EXCEPTION_CLASSES = (KeyError, IndexError, TypeError)):
        """Internal method to get data from columns within a nested set of dicts.

        Parameters
        ----------
        keys
            sequence of column name (and optionally) sample key(s) or sequence of
            list/tuple of column name, sample keys(s) which should be retrieved in
            the operation.

            Please see detailed explanation in :meth:`__getitem__()` for full
            explanation of accepted argument format / result types.

        default: Any, optional
            default value to insert in results for the case where some column
            name / sample key is not found, and the `except_missing` parameter
            is set to False.

        except_missing: bool, optional
            If False, will not throw exceptions on missing sample key value.
            Will raise KeyError if True and missing key found.

        Returns
        -------
        Any
            Single column specified, single sample key specified
        List[Any]
            arbitrary columns, multiple samples array data for each sample is
            returned in same order sample keys are received.
        """
        with ExitStack() as stack:
            if not self._is_conman:
                stack.enter_context(self)

            if isinstance(keys, str):
                return self.columns[keys]

            _COLUMNS = self._columns
            if len(keys) >= 2 and any([isinstance(k, (list, tuple)) for k in keys]):
                res = []
                for key in keys:
                    try:
                        tmp = reduce(op_getitem, key, _COLUMNS)
                        res.append(tmp)
                    except _EXCEPTION_CLASSES:
                        if except_missing:
                            raise
                        res.append(default)
                return res
            else:
                try:
                    return reduce(op_getitem, keys, _COLUMNS)
                except _EXCEPTION_CLASSES:
                    if except_missing:
                        raise
                    return default



"""Local HDF5 Backend Implementation, Identifier: ``HDF5_00``

Backend Identifiers
===================

*  Backend: ``0``
*  Version: ``0``
*  Format Code: ``00``
*  Canonical Name: ``HDF5_00``

Storage Method
==============

*  Data is written to specific subarray indexes inside an HDF5 "dataset" in a
   single HDF5 File.

*  In each HDF5 File there are ``COLLECTION_COUNT`` "datasets" (named ``["0" :
   "{COLLECTION_COUNT}"]``). These are referred to as ``"dataset number"``

*  Each dataset is a zero-initialized array of:

   *  ``dtype: {schema_dtype}``; ie ``np.float32`` or ``np.uint8``

   *  ``shape: (COLLECTION_SIZE, *{schema_shape.size})``; ie ``(500, 10)`` or
      ``(500, 300)``. The first index in the dataset is referred to as a
      ``collection index``. See technical note below for detailed explanation
      on why the flatten operaiton is performed.

*  Compression Filters, Chunking Configuration/Options are applied globally for
   all ``datasets`` in a file at dataset creation time.

*  On read and write of all samples the xxhash64_hexdigest is calculated for
   the raw array bytes. This is to ensure that all data in == data out of the
   hdf5 files. That way even if a file is manually edited (bypassing fletcher32
   filter check) we have a quick way to tell that things are not as they should
   be.

Compression Options
===================

Accepts dictionary containing keys

*  ``backend`` == ``"00"``
*  ``complib``
*  ``complevel``
*  ``shuffle``

Blosc-HDF5

*  ``complib`` valid values:

   *  ``'blosc:blosclz'``,
   *  ``'blosc:lz4'``,
   *  ``'blosc:lz4hc'``,
   *  ``'blosc:zlib'``,
   *  ``'blosc:zstd'``

*  ``complevel`` valid values: [0, 9] where 0 is "no compression" and 9 is
   "most compression"

*  ``shuffle`` valid values:

   *  ``None``
   *  ``'none'``
   *  ``'byte'``
   *  ``'bit'``


LZF Filter

*  ``'complib' == 'lzf'``
*  ``'shuffle'`` one of ``[False, None, 'none', True, 'byte']``
*  ``'complevel'`` one of ``[False, None, 'none']``

GZip Filter

*  ``'complib' == 'gzip'``
*  ``'shuffle'`` one of ``[False, None, 'none', True, 'byte']``
*  ``complevel`` valid values: [0, 9] where 0 is "no compression" and 9 is
   "most compression"


Record Format
=============

Fields Recorded for Each Array
------------------------------

*  Format Code
*  File UID
*  xxhash64_hexdigest (ie. checksum)
*  Dataset Number (``0:COLLECTION_COUNT`` dataset selection)
*  Dataset Index (``0:COLLECTION_SIZE`` dataset subarray selection)
*  Subarray Shape


Examples
--------

1)  Adding the first piece of data to a file:

    *  Array shape (Subarray Shape): (10, 10)
    *  File UID: "rlUK3C"
    *  xxhash64_hexdigest: 8067007c0f05c359
    *  Dataset Number: 16
    *  Collection Index: 105

    ``Record Data => "00:rlUK3C:8067007c0f05c359:16:105:10 10"``

1)  Adding to a piece of data to a the middle of a file:

    *  Array shape (Subarray Shape): (20, 2, 3)
    *  File UID: "rlUK3C"
    *  xxhash64_hexdigest: b89f873d3d153a9c
    *  Dataset Number: "3"
    *  Collection Index: 199

    ``Record Data => "00:rlUK3C:b89f873d3d153a9c:8:199:20 2 3"``


Technical Notes
===============

*  Files are read only after initial creation/writes. Only a write-enabled
   checkout can open a HDF5 file in ``"w"`` or ``"a"`` mode, and writer
   checkouts create new files on every checkout, and make no attempt to fill in
   unset locations in previous files. This is not an issue as no disk space is
   used until data is written to the initially created "zero-initialized"
   collection datasets

*  On write: Single Writer Multiple Reader (``SWMR``) mode is set to ensure that
   improper closing (not calling ``.close()``) method does not corrupt any data
   which had been previously flushed to the file.

*  On read: SWMR is set to allow multiple readers (in different threads /
   processes) to read from the same file. File handle serialization is handled
   via custom python ``pickle`` serialization/reduction logic which is
   implemented by the high level ``pickle`` reduction ``__set_state__()``,
   ``__get_state__()`` class methods.

*  An optimization is performed in order to increase the read / write
   performance of variable shaped datasets. Due to the way that we initialize
   an entire HDF5 file with all datasets pre-created (to the size of the max
   subarray shape), we need to ensure that storing smaller sized arrays (in a
   variable sized Hangar Column) would be effective. Because we use chunked
   storage, certain dimensions which are incomplete could have potentially
   required writes to chunks which do are primarily empty (worst case "C" index
   ordering), increasing read / write speeds significantly.

   To overcome this, we create HDF5 datasets which have ``COLLECTION_SIZE``
   first dimension size, and only ONE second dimension of size
   ``schema_shape.size()`` (ie. product of all dimensions). For example an
   array schema with shape (10, 10, 3) would be stored in a HDF5 dataset of
   shape (COLLECTION_SIZE, 300). Chunk sizes are chosen to align on the first
   dimension with a second dimension of size which fits the total data into L2
   CPU Cache (< 256 KB). On write, we use the ``np.ravel`` function to
   construct a "view" (not copy) of the array as a 1D array, and then on read
   we reshape the array to the recorded size (a copyless "view-only"
   operation). This is part of the reason that we only accept C ordered arrays
   as input to Hangar.
"""
import logging
import os
from collections import ChainMap
from contextlib import suppress
from functools import partial
from pathlib import Path
from typing import MutableMapping, Tuple, Optional, Union, Callable

import h5py
import numpy as np

try:
    # hdf5plugin warns if a filter is already loaded.
    _logger = logging.getLogger('hdf5plugin')
    _initialLevel = _logger.getEffectiveLevel()
    _logger.setLevel(logging.ERROR)
    import hdf5plugin
    if 'blosc' not in hdf5plugin.FILTERS:
        raise ImportError(f'BLOSC unavailable via hdf5plugin: {hdf5plugin.FILTERS}')
finally:
    _logger.setLevel(_initialLevel)
from xxhash import xxh64_hexdigest

from .specs import HDF5_00_DataHashSpec
from .. import __version__
from ..optimized_utils import SizedDict
from ..constants import DIR_DATA_REMOTE, DIR_DATA_STAGE, DIR_DATA_STORE, DIR_DATA
from ..utils import find_next_prime, random_string, set_blosc_nthreads
from ..op_state import reader_checkout_only, writer_checkout_only
from ..typesystem import Descriptor, OneOf, DictItems, SizedIntegerTuple, checkedmeta

set_blosc_nthreads()

# ----------------------------- Configuration ---------------------------------

_FmtCode = '00'

# contents of a single hdf5 file
COLLECTION_SIZE = 250
COLLECTION_COUNT = 100

# chunking options for compression schemes
CHUNK_MAX_NBYTES = 255_000  # < 256 KB to fit in L2 CPU Cache
CHUNK_MAX_RDCC_NBYTES = 100_000_000
CHUNK_RDCC_W0 = 0.75

# -------------------------------- Parser Implementation ----------------------


def hdf5_00_encode(uid: str, cksum: str, dset: int, dset_idx: int, shape: Tuple[int]) -> bytes:
    """converts the hdf5 data has spec to an appropriate db value

    Parameters
    ----------
    uid : str
        the file name prefix which the data is written to.
    cksum : int
        xxhash_64.hex_digest checksum of the data bytes in numpy array form.
    dset : int
        collection (ie. hdf5 dataset) name to find this data piece.
    dset_idx : int
        collection first axis index in which this data piece resides.
    shape : Tuple[int]
        shape of the data sample written to the collection idx. ie:
        what subslices of the hdf5 dataset should be read to retrieve
        the sample as recorded.

    Returns
    -------
    bytes
        hash data db value recording all input specifications.
    """
    shape_str = " ".join([str(i) for i in shape])
    return f'00:{uid}:{cksum}:{dset}:{dset_idx}:{shape_str}'.encode()


# ------------------------- Accessor Object -----------------------------------


@DictItems(
    expected_keys_required={'complib': True, 'complevel': True, 'shuffle': True},
    expected_values={
        'complib': ['blosc:blosclz', 'blosc:lz4','blosc:lz4hc', 'blosc:zlib', 'blosc:zstd'],
        'complevel': [i for i in range(10)],
        'shuffle': [None, 'none', 'byte', 'bit']})
class BloscCompressionOptions(Descriptor):
    pass


@DictItems(
    expected_keys_required={'complib': True, 'complevel': True, 'shuffle': True},
    expected_values={
        'complib': ['gzip'], 'complevel': [i for i in range(10)], 'shuffle': [True, False]})
class GzipCompressionOptions(Descriptor):
    pass


@DictItems(
    expected_keys_required={'complib': True, 'complevel': False, 'shuffle': True},
    expected_values={
        'complib': ['lzf'], 'complevel': ['none', None], 'shuffle': [True, False]})
class LzfCompressionOptions(Descriptor):
    pass


@OneOf(list(map(lambda x: np.dtype(x).name, [
        np.bool, np.uint8, np.uint16, np.uint32, np.uint64, np.int8, np.int16,
        np.int32, np.int64, np.float16, np.float32, np.float64, np.longdouble])))
class AllowedDtypes(Descriptor):
    """
    Note. np.longdouble since np.float128 not guaranteed to be available on
    all system. this is a particular issue with some windows numpy builds
    """
    pass


class HDF5_00_Options(metaclass=checkedmeta):
    _shape = SizedIntegerTuple(size=32)
    _dtype = AllowedDtypes()
    _lzf = LzfCompressionOptions()
    _gzip = GzipCompressionOptions()
    _blosc = BloscCompressionOptions()
    _avail_filters = ('_lzf', '_gzip', '_blosc')

    def __init__(self, backend_options, dtype, shape, *args, **kwargs):
        self._shape = shape
        self._dtype = dtype
        self._selected_filter = None
        if backend_options is None:
            backend_options = self.default_options

        for filter_attr in self._avail_filters:
            with suppress((KeyError, ValueError)):
                setattr(self, filter_attr, backend_options)
                self._selected_filter = filter_attr
                break
        else:  # N.B. for-else loop (ie. "no-break")
            raise ValueError(f'Invalid backend_options {backend_options}')
        self._verify_data_nbytes_larger_than_clib_min()

    def _verify_data_nbytes_larger_than_clib_min(self):
        """blosc clib should not be used if data buffer size < 16 bytes.

        Raises
        ------
        ValueError:
            if the data size is not valid for the clib
        """
        if self._selected_filter in ['_blosc', None]:
            num_items = np.prod(self._shape)
            itemsize = np.dtype(self._dtype).itemsize
            nbytes = itemsize * num_items
            if nbytes <= 16:
                raise ValueError(f'blosc clib requires data buffer size > 16 bytes')

    @property
    def default_options(self):
        if 'blosc' in hdf5plugin.FILTERS:
            try:
                self._verify_data_nbytes_larger_than_clib_min()
                return {'complib': 'blosc:lz4hc', 'complevel': 5, 'shuffle': 'byte'}
            except ValueError:
                pass
        return {'complib': 'lzf', 'complevel': None, 'shuffle': True}

    @property
    def backend_options(self):
        return getattr(self, self._selected_filter)

    @property
    def init_requires(self):
        return ('repo_path', 'schema_shape', 'schema_dtype')


HDF5_00_MapTypes = MutableMapping[str, Union[h5py.File, Callable[[], h5py.File]]]


class HDF5_00_FileHandles(object):
    """Manage HDF5 file handles.

    When in SWMR-write mode, no more than a single file handle can be in the
    "writeable" state. This is an issue where multiple columns may need to
    write to the same column schema.
    """

    def __init__(self, repo_path: Path, schema_shape: tuple, schema_dtype: np.dtype):
        self.path: Path = repo_path
        self.schema_shape: tuple = schema_shape
        self.schema_dtype: np.dtype = schema_dtype
        self._dflt_backend_opts: Optional[dict] = None

        self.rFp: HDF5_00_MapTypes = {}
        self.wFp: HDF5_00_MapTypes = {}
        self.Fp: HDF5_00_MapTypes = ChainMap(self.rFp, self.wFp)
        self.rDatasets = SizedDict(maxsize=100)
        self.wdset: Optional[h5py.Dataset] = None

        self.mode: Optional[str] = None
        self.hIdx: Optional[int] = None
        self.w_uid: Optional[str] = None
        self.hMaxSize: Optional[int] = None
        self.hNextPath: Optional[int] = None
        self.hColsRemain: Optional[int] = None

        self.STAGEDIR: Path = Path(self.path, DIR_DATA_STAGE, _FmtCode)
        self.REMOTEDIR: Path = Path(self.path, DIR_DATA_REMOTE, _FmtCode)
        self.STOREDIR: Path = Path(self.path, DIR_DATA_STORE, _FmtCode)
        self.DATADIR: Path = Path(self.path, DIR_DATA, _FmtCode)
        self.DATADIR.mkdir(exist_ok=True)

    def __enter__(self):
        return self

    def __exit__(self, *exc):
        if self.w_uid in self.wFp:
            self.wFp[self.w_uid]['/'].attrs.modify('next_location', (self.hNextPath, self.hIdx))
            self.wFp[self.w_uid]['/'].attrs.modify('collections_remaining', self.hColsRemain)
            self.wFp[self.w_uid].flush()

    @reader_checkout_only
    def __getstate__(self) -> dict:
        """ensure multiprocess operations can pickle relevant data.
        """
        self.close()
        state = self.__dict__.copy()
        del state['rFp']
        del state['wFp']
        del state['Fp']
        del state['rDatasets']
        del state['wdset']
        return state

    def __setstate__(self, state: dict) -> None:  # pragma: no cover
        """ensure multiprocess operations can pickle relevant data.
        """
        self.__dict__.update(state)
        self.rFp = {}
        self.wFp = {}
        self.Fp = ChainMap(self.rFp, self.wFp)
        self.rDatasets = {}
        self.wdset = None
        self.open(mode=self.mode)

    @property
    def backend_opts(self):
        return self._dflt_backend_opts

    @writer_checkout_only
    def _backend_opts_set(self, val):
        """Nonstandard descriptor method. See notes in ``backend_opts.setter``.
        """
        self._dflt_backend_opts = val
        return

    @backend_opts.setter
    def backend_opts(self, value):
        """
        Using seperate setter method (with ``@writer_checkout_only`` decorator
        applied) due to bug in python <3.8.

        From: https://bugs.python.org/issue19072
            > The classmethod decorator when applied to a function of a class,
            > does not honour the descriptor binding protocol for whatever it
            > wraps. This means it will fail when applied around a function which
            > has a decorator already applied to it and where that decorator
            > expects that the descriptor binding protocol is executed in order
            > to properly bind the function to the class.
        """
        return self._backend_opts_set(value)

    def open(self, mode: str, *, remote_operation: bool = False):
        """Open an hdf5 file handle in the Handler Singleton

        Parameters
        ----------
        mode : str
            one of `r` or `a` for read only / read-write.
        remote_operation : optional, kwarg only, bool
            if this hdf5 data is being created from a remote fetch operation, then
            we don't open any files for reading, and only open files for writing
            which exist in the remote data dir. (default is false, which means that
            write operations use the stage data dir and read operations use data store
            dir)
        """
        self.mode = mode
        if self.mode == 'a':
            process_dir = self.REMOTEDIR if remote_operation else self.STAGEDIR
            process_dir.mkdir(exist_ok=True)
            for uidpth in process_dir.iterdir():
                if uidpth.suffix == '.hdf5':
                    file_pth = self.DATADIR.joinpath(uidpth.name)
                    self.rFp[uidpth.stem] = partial(
                        h5py.File, file_pth, 'r', swmr=True, libver='latest')

        if not remote_operation:
            if not self.STOREDIR.is_dir():
                return
            for uidpth in self.STOREDIR.iterdir():
                if uidpth.suffix == '.hdf5':
                    file_pth = self.DATADIR.joinpath(uidpth.name)
                    self.rFp[uidpth.stem] = partial(
                        h5py.File, file_pth, 'r', swmr=True, libver='latest')

    def close(self):
        """Close a file handle after writes have been completed

        behavior changes depending on write-enable or read-only file

        Returns
        -------
        bool
            True if success, otherwise False.
        """
        if self.mode == 'a':
            if self.w_uid in self.wFp:
                self.wFp[self.w_uid]['/'].attrs.modify('next_location', (self.hNextPath, self.hIdx))
                self.wFp[self.w_uid]['/'].attrs.modify('collections_remaining', self.hColsRemain)
                self.wFp[self.w_uid].flush()
            for uid in list(self.wFp.keys()):
                with suppress(AttributeError):
                    self.wFp[uid].close()
                del self.wFp[uid]
            self.wdset = None
            self.hMaxSize = None
            self.hNextPath = None
            self.hIdx = None
            self.hColsRemain = None
            self.w_uid = None

        for uid in list(self.rFp.keys()):
            with suppress(AttributeError):
                self.rFp[uid].close()
            del self.rFp[uid]
        self.rDatasets = {}

    @staticmethod
    def delete_in_process_data(repo_path: Path, *, remote_operation=False) -> None:
        """Removes some set of files entirely from the stage/remote directory.

        DANGER ZONE. This should essentially only be used to perform hard resets
        of the repository state.

        Parameters
        ----------
        repo_path : Path
            path to the repository on disk
        remote_operation : optional, kwarg only, bool
            If true, modify contents of the remote_dir, if false (default) modify
            contents of the staging directory.
        """
        data_dir = Path(repo_path, DIR_DATA, _FmtCode)
        PDIR = DIR_DATA_STAGE if not remote_operation else DIR_DATA_REMOTE
        process_dir = Path(repo_path, PDIR, _FmtCode)
        if not process_dir.is_dir():
            return

        for uidpth in process_dir.iterdir():
            if uidpth.suffix == '.hdf5':
                os.remove(process_dir.joinpath(uidpth.name))
                os.remove(data_dir.joinpath(uidpth.name))
        os.rmdir(process_dir)

    @staticmethod
    def _dataset_opts(complib: str, complevel: int, shuffle: Union[bool, str]) -> dict:
        """specify compression options for the hdf5 dataset.

        .. seealso:: :function:`_blosc_opts`

        to enable blosc compression, use the conda-forge `blosc-hdf5-plugin` package.

        .. seealso::

        * https://github.com/conda-forge/staged-recipes/pull/7650
        * https://github.com/h5py/h5py/issues/611

        Parameters
        ----------
        complib : str
            the compression lib to use, one of ['lzf', 'gzip', 'blosc:blosclz',
            'blosc:lz4', 'blosc:lz4hc', 'blosc:zlib', 'blosc:zstd']
        complevel : int
            compression level to specify (accepts values [0, 9] for all except 'lzf'
            where no complevel is accepted)
        shuffle : bool
            if True or `byte`, enable byte shuffle filter, if blosc
            compression, pass through 'bits' is accepted as well. False, or
            None indicates no shuffle should be applied.
        """
        # ---- blosc hdf5 plugin filters ----
        _blosc_compression = {
            'blosc:blosclz': 0,
            'blosc:lz4': 1,
            'blosc:lz4hc': 2,
            # Not built 'snappy': 3,
            'blosc:zlib': 4,
            'blosc:zstd': 5}
        _blosc_shuffle = {None: 0, 'none': 0, 'byte': 1, 'bit': 2}
        _blosc_complevel = {**{i: i for i in range(10)}, None: 9, 'none': 9}

        # ---- h5py built in filters ----
        _lzf_gzip_shuffle = {None: False, False: False, 'none': False, True: True, 'byte': True}
        _lzf_complevel = {False: None, None: None, 'none': None}
        _gzip_complevel = {**{i: i for i in range(10)}, None: 4, 'none': 4}

        if complib.startswith('blosc'):
            args = {
                'compression': 32001,
                'compression_opts': (
                    0, 0, 0, 0,
                    _blosc_complevel[complevel],
                    _blosc_shuffle[shuffle],
                    _blosc_compression[complib]),
                'shuffle': False}
        elif complib == 'lzf':
            args = {
                'shuffle': _lzf_gzip_shuffle[shuffle],
                'compression': complib,
                'compression_opts': _lzf_complevel[complevel]}
        elif complib == 'gzip':
            args = {
                'shuffle': _lzf_gzip_shuffle[shuffle],
                'compression': complib,
                'compression_opts': _gzip_complevel[complevel]}
        elif complib in (None, False, 'none'):
            args = {
                'shuffle': False,
                'compression': None,
                'compression_opts': None}
        else:
            raise ValueError(f'unknown value for opt arg `complib`: {complib}')
        return args

    @staticmethod
    def _chunk_opts(sample_array: np.ndarray, max_chunk_nbytes: int) -> Tuple[list, int]:
        """Determine the chunk shape so each array chunk fits into configured nbytes.

        Currently the chunk nbytes are not user configurable. Instead the constant
        `HDF5_MAX_CHUNK_NBYTES` is sued to determine when to split.

        Parameters
        ----------
        sample_array : `np.array`
            Sample array whose shape and dtype should be used as the basis of the
            chunk shape determination
        max_chunk_nbytes : int
            how many bytes the array chunks should be limited to.

        Returns
        -------
        list
            list of ints of length == rank of `sample_array` specifying chunk sizes
            to split `sample_array` into nbytes
        int
            nbytes which the chunk will fit in. Will be <= `HDF5_MAX_CHUNK_NBYTES`
        """
        chunk_size = int(np.floor(max_chunk_nbytes / sample_array.itemsize))
        if chunk_size > sample_array.size:
            chunk_size = sample_array.size
        chunk_shape = [chunk_size]
        chunk_nbytes = np.zeros(shape=chunk_shape, dtype=sample_array.dtype).nbytes

        return (chunk_shape, chunk_nbytes)

    def _create_schema(self, *, remote_operation: bool = False):
        """stores the shape and dtype as the schema of a column.

        Parameters
        ----------
        remote_operation : optional, kwarg only, bool
            if this schema is being created from a remote fetch operation, then do not
            place the file symlink in the staging directory. Instead symlink it
            to a special remote staging directory. (default is False, which places the
            symlink in the stage data directory.)

        Notes
        -----

        Parameters set for raw-data-chunk-cache (rdcc) values:

        * rdcc_nbytes: sets the total size (measured in bytes) of the raw data chunk
          cache for each dataset. This should be set to the size of each chunk times
          the number of chunks that are likely to be needed in cache.
        * rdcc_w0: sets the policy for chunks to be removed from the cache when more
          space is needed. If set to 0, always evict the least recently used chunk in
          cache. If set to 1, always evict the least recently used chunk which has
          been fully read or written. If the value is between 0 and 1, the behavior
          will be a blend of the two.
        * rdcc_nslots: The number of chunk slots in the cache for this entire file.
          In order for quick lookup, a hash map is used for each chunk value. For
          maximum performance, this value should be set approximately 100 times that
          number of chunks.

        .. seealso::

            http://docs.h5py.org/en/stable/high/file.html#chunk-cache

        """
        # -------------------- Chunk & RDCC Vals ------------------------------

        sample_array = np.zeros(self.schema_shape, dtype=self.schema_dtype)
        chunk_shape, chunk_nbytes = self._chunk_opts(
            sample_array=sample_array, max_chunk_nbytes=CHUNK_MAX_NBYTES)

        rdcc_nbytes_val = sample_array.nbytes * COLLECTION_SIZE
        if rdcc_nbytes_val < CHUNK_MAX_NBYTES:
            rdcc_nbytes_val = CHUNK_MAX_NBYTES
        elif rdcc_nbytes_val > CHUNK_MAX_RDCC_NBYTES:
            rdcc_nbytes_val = CHUNK_MAX_RDCC_NBYTES

        rdcc_nslots_guess = np.math.ceil(rdcc_nbytes_val / chunk_nbytes) * 100
        rdcc_nslots_prime_val = find_next_prime(rdcc_nslots_guess)

        # ---------------------------- File Creation --------------------------

        uid = random_string()
        file_path = self.DATADIR.joinpath(f'{uid}.hdf5')
        self.wFp[uid] = h5py.File(file_path,
                                  mode='w',
                                  libver='latest',
                                  rdcc_nbytes=rdcc_nbytes_val,
                                  rdcc_w0=CHUNK_RDCC_W0,
                                  rdcc_nslots=rdcc_nslots_prime_val)
        self.w_uid = uid
        self.wdset = None
        self.hNextPath = 0
        self.hIdx = 0
        self.hColsRemain = COLLECTION_COUNT
        self.hMaxSize = COLLECTION_SIZE

        process_dir = self.REMOTEDIR if remote_operation else self.STAGEDIR
        Path(process_dir, f'{uid}.hdf5').touch()

        # ----------------------- Dataset Creation ----------------------------

        optKwargs = self._dataset_opts(**self._dflt_backend_opts)
        for dset_num in range(COLLECTION_COUNT):
            self.wFp[uid].create_dataset(
                f'/{dset_num}',
                shape=(COLLECTION_SIZE, sample_array.size),
                dtype=sample_array.dtype,
                maxshape=(COLLECTION_SIZE, sample_array.size),
                chunks=(1, *chunk_shape),
                **optKwargs)

        # ---------------------- Attribute Config Vals ------------------------

        self.wFp[self.w_uid]['/'].attrs['HANGAR_VERSION'] = __version__
        self.wFp[self.w_uid]['/'].attrs['schema_shape'] = sample_array.shape
        self.wFp[self.w_uid]['/'].attrs['schema_dtype_num'] = sample_array.dtype.num
        self.wFp[self.w_uid]['/'].attrs['next_location'] = (0, 0)
        self.wFp[self.w_uid]['/'].attrs['collection_max_size'] = COLLECTION_SIZE
        self.wFp[self.w_uid]['/'].attrs['collection_total'] = COLLECTION_COUNT
        self.wFp[self.w_uid]['/'].attrs['collections_remaining'] = COLLECTION_COUNT
        self.wFp[self.w_uid]['/'].attrs['rdcc_nbytes'] = rdcc_nbytes_val
        self.wFp[self.w_uid]['/'].attrs['rdcc_w0'] = CHUNK_RDCC_W0
        self.wFp[self.w_uid]['/'].attrs['rdcc_nslots'] = rdcc_nslots_prime_val
        self.wFp[self.w_uid]['/'].attrs['chunk_shape'] = chunk_shape
        if optKwargs['compression_opts'] is not None:
            self.wFp[self.w_uid]['/'].attrs['compression_opts'] = optKwargs['compression_opts']
        else:
            self.wFp[self.w_uid]['/'].attrs['compression_opts'] = False

        self.wFp[self.w_uid].flush()
        try:
            self.wFp[self.w_uid].swmr_mode = True
        except ValueError:
            assert self.wFp[self.w_uid].swmr_mode is True
        self.wdset = self.wFp[self.w_uid][f'/{self.hNextPath}']

    def read_data(self, hashVal: HDF5_00_DataHashSpec) -> np.ndarray:
        """Read data from an hdf5 file handle at the specified locations

        Parameters
        ----------
        hashVal : HDF5_00_DataHashSpec
            record specification parsed from its serialized store val in lmdb.

        Returns
        -------
        np.array
            requested data.
        """
        arrSize = 1
        for dim in hashVal.shape:
            arrSize *= dim
        srcSlc = (hashVal.dataset_idx, slice(0, arrSize))
        dsetCol = f'/{hashVal.dataset}'
        rdictkey = f'{hashVal.uid}{dsetCol}'

        if self.schema_dtype:  # if is not None
            destArr = np.empty((arrSize,), self.schema_dtype)
            if rdictkey in self.rDatasets:
                self.rDatasets[rdictkey].read_direct(destArr, srcSlc, None)
            else:
                try:
                    self.Fp[hashVal.uid][dsetCol].read_direct(destArr, srcSlc, None)
                    self.rDatasets[rdictkey] = self.Fp[hashVal.uid][dsetCol]
                except TypeError:
                    self.Fp[hashVal.uid] = self.Fp[hashVal.uid]()
                    self.rDatasets[rdictkey] = self.Fp[hashVal.uid][dsetCol]
                    self.rDatasets[rdictkey].read_direct(destArr, srcSlc, None)
                except KeyError:
                    process_dir = self.STAGEDIR if self.mode == 'a' else self.STOREDIR
                    if Path(process_dir, f'{hashVal.uid}.hdf5').is_file():
                        file_pth = self.DATADIR.joinpath(f'{hashVal.uid}.hdf5')
                        self.rFp[hashVal.uid] = h5py.File(file_pth, 'r', swmr=True, libver='latest')
                        self.rDatasets[rdictkey] = self.Fp[hashVal.uid][dsetCol]
                        self.rDatasets[rdictkey].read_direct(destArr, srcSlc, None)
                    else:
                        raise
        else:
            if rdictkey in self.rDatasets:
                destArr = self.rDatasets[rdictkey][srcSlc]
            else:
                try:
                    destArr = self.Fp[hashVal.uid][dsetCol][srcSlc]
                    self.rDatasets[rdictkey] = self.Fp[hashVal.uid][dsetCol]
                except TypeError:
                    self.Fp[hashVal.uid] = self.Fp[hashVal.uid]()
                    destArr = self.Fp[hashVal.uid][dsetCol][srcSlc]
                    self.rDatasets[rdictkey] = self.Fp[hashVal.uid][dsetCol]
                except KeyError:
                    process_dir = self.STAGEDIR if self.mode == 'a' else self.STOREDIR
                    if Path(process_dir, f'{hashVal.uid}.hdf5').is_file():
                        file_pth = self.DATADIR.joinpath(f'{hashVal.uid}.hdf5')
                        self.rFp[hashVal.uid] = h5py.File(file_pth, 'r', swmr=True, libver='latest')
                        destArr = self.Fp[hashVal.uid][dsetCol][srcSlc]
                        self.rDatasets[rdictkey] = self.Fp[hashVal.uid][dsetCol]
                    else:
                        raise

        out = destArr.reshape(hashVal.shape)
        if xxh64_hexdigest(out) != hashVal.checksum:
            # try casting to check if dtype does not match for all zeros case
            out = out.astype(np.typeDict[self.Fp[hashVal.uid]['/'].attrs['schema_dtype_num']])
            if xxh64_hexdigest(out) != hashVal.checksum:
                raise RuntimeError(
                    f'DATA CORRUPTION Checksum {xxh64_hexdigest(out)} != recorded {hashVal}')
        return out

    def write_data(self, array: np.ndarray, *, remote_operation: bool = False) -> bytes:
        """verifies correctness of array data and performs write operation.

        Parameters
        ----------
        array : np.ndarray
            tensor to write to group.
        remote_operation : optional, kwarg only, bool
            If this is a remote process which is adding data, any necessary
            hdf5 dataset files will be created in the remote data dir instead
            of the stage directory. (default is False, which is for a regular
            access process)

        Returns
        -------
        bytes
            string identifying the collection dataset and collection dim-0 index
            which the array can be accessed at.
        """
        checksum = xxh64_hexdigest(array)
        if self.w_uid in self.wFp:
            self.hIdx += 1
            if self.hIdx >= self.hMaxSize:
                self.wdset.flush()
                self.hIdx = 0
                self.hNextPath += 1
                self.hColsRemain -= 1
                self.wdset = self.wFp[self.w_uid][f'/{self.hNextPath}']
                if self.hColsRemain <= 1:
                    self.wFp[self.w_uid]['/'].attrs.modify('next_location', (self.hNextPath, self.hIdx))
                    self.wFp[self.w_uid]['/'].attrs.modify('collections_remaining', self.hColsRemain)
                    self.wFp[self.w_uid].flush()
                    self._create_schema(remote_operation=remote_operation)
        else:
            self._create_schema(remote_operation=remote_operation)

        destSlc = (self.hIdx, slice(0, array.size))
        flat_arr = np.ravel(array)
        self.wdset.write_direct(flat_arr, None, destSlc)
        self.wdset.flush()
        return hdf5_00_encode(self.w_uid, checksum, self.hNextPath, self.hIdx, array.shape)



import click


class StrOrIntType(click.ParamType):
    """Custom type for click to parse the sample name
    argument to integer or string
    """

    def convert(self, value, param, ctx):
        if not value:
            return None

        try:
            stype, sample = value.split(':') if ':' in value else ('str', value)
        except ValueError:
            self.fail(f"Sample name {value} not formatted properly", param, ctx)
        try:
            if stype not in ('str', 'int'):
                self.fail(f"type {stype} is not allowed", param, ctx)
            return int(sample) if stype == 'int' else str(sample)
        except (ValueError, TypeError):
            self.fail(f"{sample} is not a valid {stype}", param, ctx)


def parse_custom_arguments(click_args: list) -> dict:
    """
    Parse all the unknown arguments from click for downstream tasks. Used in
    user plugins for custom command line arguments.

    Parameters
    ----------
    click_args : list
        Unknown arguments from click

    Returns
    -------
    parsed : dict
        Parsed arguments stored as key value pair

    Note
    -----
    Unknown arguments must be long arguments i.e should start with --
    """
    parsed = {}
    for i in range(0, len(click_args), 2):
        key = click_args[i]
        val = click_args[i + 1]
        if not key.startswith('--'):
            raise RuntimeError(f"Could not parse argument {key}. It should be prefixed with `--`")
        parsed[key[2:]] = val
    return parsed



from .checkout_iteration import CheckoutDictIteration
from .datasetget import GetMixin
from .recorditer import CursorRangeIterator

__all__ = ['GetMixin', 'CursorRangeIterator', 'CheckoutDictIteration']



import warnings
from collections import defaultdict
from typing import NamedTuple

import lmdb

from .parsing import (
    remote_db_key_from_raw_key,
    remote_db_val_from_raw_val,
    remote_raw_key_from_db_key,
    remote_raw_val_from_db_val,
    repo_branch_head_db_key_from_raw_key,
    repo_branch_head_db_val_from_raw_val,
    repo_branch_head_raw_key_from_db_key,
    repo_branch_head_raw_val_from_db_val,
    repo_head_db_key,
    repo_head_db_val_from_raw_val,
    repo_head_raw_val_from_db_val,
    repo_writer_lock_db_key,
    repo_writer_lock_db_val_from_raw_val,
    repo_writer_lock_force_release_sentinal,
    repo_writer_lock_sentinal_db_val,
)
from ..constants import K_REMOTES, K_BRANCH
from ..txnctx import TxnRegister


class BranchHead(NamedTuple):
    name: str
    digest: str


"""
Write operation enabled lock methods
------------------------------------

Any operation which wants to interact with the main storage services in a
write-enabled way must acquire a lock to perform the operation. See docstrings
below for more info
"""


def writer_lock_held(branchenv):
    """Check to see if the writer lock is free before attempting to acquire it.

    Parameters
    ----------
    branchenv : lmdb.Environment
        lmdb environment where the writer lock is stored

    Returns
    -------
    bool
        True if the lock is available to take, False if it is currently held.
    """
    writerLockKey = repo_writer_lock_db_key()
    writerLockSentinalVal = repo_writer_lock_sentinal_db_val()
    branchtxn = TxnRegister().begin_reader_txn(branchenv)
    try:
        currentWriterLockVal = branchtxn.get(writerLockKey)
        if currentWriterLockVal == writerLockSentinalVal:
            lockAvailable = True
        elif currentWriterLockVal is None:
            # on first initialization, writer lock key/val is not set.
            lockAvailable = True
        else:
            lockAvailable = False
    finally:
        TxnRegister().abort_reader_txn(branchenv)
    return lockAvailable


def acquire_writer_lock(branchenv, writer_uuid):
    """Attempt to acquire the writer lock for a write-enabled checkout object.

    If the writer_uuid matches the recorded value, or the lock is available (or
    uninitialized entirely in the case of a brand-new repository), the lock will
    be updated with the requested uuid, and no other write-enabled checkout can
    be started until it is either released, or a force reset is performed (in
    the event of a system crash or user error.)

    Parameters
    ----------
    branchenv : lmdb.Environment
        lmdb environment where the writer lock is stored
    writer_uuid : str
        uuid generated when a write enabled checkout instance starts

    Returns
    -------
    bool
        success of the operation, which will be validated by the writer class as
        a safety net incase the upstream in the event some user code tries to
        catch the exception.Z

    Raises
    ------
    PermissionError
        If the lock can not be acquired

    """
    writerLockKey = repo_writer_lock_db_key()
    writerLockSentinalVal = repo_writer_lock_sentinal_db_val()
    requestWriterLockVal = repo_writer_lock_db_val_from_raw_val(writer_uuid)

    branchtxn = TxnRegister().begin_writer_txn(branchenv)
    try:
        currentWriterLockVal = branchtxn.get(writerLockKey)
        if currentWriterLockVal == requestWriterLockVal:
            success = True
        elif currentWriterLockVal == writerLockSentinalVal:
            branchtxn.put(writerLockKey, requestWriterLockVal)
            success = True
        elif currentWriterLockVal is None:
            # on first initialization, writer lock key/val is not set.
            branchtxn.put(writerLockKey, requestWriterLockVal)
            success = True
        else:
            err = 'Cannot acquire the writer lock. Only one instance of a writer checkout '\
                  'can be active at a time. If the last checkout of this repository did '\
                  'not properly close, or a crash occurred, the lock must be manually freed '\
                  'before another writer can be instantiated.'
            raise PermissionError(err)
    finally:
        TxnRegister().commit_writer_txn(branchenv)

    return success


def release_writer_lock(branchenv, writer_uuid):
    """Internal method to release a writer lock held by a specified uuid.

    This method also accept the force-release sentinel by a caller in the
    writer_uuid field. If the writer_uuid does not match the lock value (and the
    force sentinel is not used), then a runtime error will be thrown and no-op
    performed

    Parameters
    ----------
    branchenv : lmdb.Environment
        lmdb environment where the lock key/val lives
    writer_uuid : str
        uuid of the requested releaser

    Returns
    -------
    bool
        if the operation was successful or now

    Raises
    ------
    RuntimeError
        if the request uuid does not match the lock value.
    """
    writerLockKey = repo_writer_lock_db_key()
    forceReleaseSentinal = repo_writer_lock_force_release_sentinal()
    lockSentinalVal = repo_writer_lock_sentinal_db_val()
    requestWriterLockVal = repo_writer_lock_db_val_from_raw_val(writer_uuid)

    txn = TxnRegister().begin_writer_txn(branchenv)
    try:
        currentLockVal = txn.get(writerLockKey)
        if writer_uuid == forceReleaseSentinal:
            warnings.warn('Writer lock force successfully force released.', ResourceWarning)
            txn.put(writerLockKey, lockSentinalVal)
            success = True
        elif currentLockVal == requestWriterLockVal:
            txn.put(writerLockKey, lockSentinalVal)
            success = True
        elif currentLockVal == lockSentinalVal:
            warnings.warn('The lock is already available, no release is necessary.', UserWarning)
            success = True
        else:
            err = f'FATAL ERROR Requested release of writer lock: {currentLockVal} by '\
                  f'non-valid requestor: {requestWriterLockVal} -- How did this happen?'
            raise RuntimeError(err)
    finally:
        TxnRegister().commit_writer_txn(branchenv)

    return success


"""
Methods to interact with the branch head records
------------------------------------------------

.. todo::
   Need a delete branch operation.
"""

# ---------------- branch creation and deletion operations ------------------------------


def create_branch(branchenv, name, base_commit) -> BranchHead:
    """Internal operations used to create a branch.

    Parameters
    ----------
    branchenv : lmdb.Environment
        lmdb environment of the branch db
    name : str
        Name of the branch to create, if a branch with this name exists no
        operation  will occur and a `ValueError` will be thrown.
    base_commit : str
        The commit to start this branch from.

    Returns
    -------
    BranchHead
        NamedTuple[str, str] with fields for `name` and `digest` of the branch
        created (if the operation was successful)

    Raises
    ------
    ValueError
        If the branch already exists, no-op and raise this.
    RuntimeError
        If the repository does not have at-least one commit on the `default`
        (ie. `master`) branch.
    """
    if base_commit is None:
        headBranch = get_staging_branch_head(branchenv)
        base_commit = get_branch_head_commit(branchenv, headBranch)
        if (headBranch == 'master') and (base_commit == ''):
            msg = 'At least one commit must be made in the repository on the `default` '\
                  '(`master`) branch before new branches can be created'
            raise RuntimeError(msg)

    branchHeadKey = repo_branch_head_db_key_from_raw_key(name)
    branchHeadVal = repo_branch_head_db_val_from_raw_val(base_commit)

    branchtxn = TxnRegister().begin_writer_txn(branchenv)
    try:
        success = branchtxn.put(branchHeadKey, branchHeadVal, overwrite=False)
        if success is False:
            err = f'A branch with the name {name} already exists, please specify'\
                  f'a different name or delete the branch.'
            raise ValueError(err)
    finally:
        TxnRegister().commit_writer_txn(branchenv)

    return BranchHead(name=name, digest=base_commit)


def remove_branch(branchenv: lmdb.Environment,
                  refenv: lmdb.Environment,
                  name: str,
                  *,
                  force_delete: bool = False) -> BranchHead:
    """Remove a branch head pointer after verifying validity and safety

    Parameters
    ----------
    branchenv : lmdb.Environment
        db containing the branch head specs
    refenv : lmdb.Environment
        db containing the commit refs
    name : str
        name of the branch which should be deleted.
    force_delete : bool, optional
        If True, remove the branch pointer even if the changes are un-merged in
        other branch histories. by default False

    Returns
    -------
    BranchHead
        NamedTuple[str, str] with fields for `name` and `digest` of the branch
        pointer deleted.

    Raises
    ------
    ValueError
        If a branch with the provided name does not exist locally
    PermissionError
        If removal of the branch would result in a repository with zero local
        branches.
    PermissionError
        If a write enabled checkout is holding the writer-lock at time of this
        call.
    PermissionError
        If the branch to be removed was the last used in a write-enabled
        checkout, and whose contents form the base of the staging area.
    RuntimeError
        If the branch has not been fully merged into other branch histories,
        and ``force_delete`` option is not ``True``.
    """
    from .commiting import get_commit_ancestors_graph

    all_branches = get_branch_names(branchenv)
    alive_branches = [x for x in all_branches if '/' not in x]  # exclude remotes
    if name not in alive_branches:
        raise ValueError(f'Branch: {name} does not exist')

    alive_branches.remove(name)
    if len(alive_branches) == 0:
        msg = f'Not allowed to remove all branches from a repository! '\
              f'Operation aborted without completing removal of branch: {name}'
        raise PermissionError(msg)

    if writer_lock_held(branchenv) is False:
        msg = f'Cannot remove branch when a `write-enabled` checkout is active. '\
              f're-run after committing/closing the writer.'
        raise PermissionError(msg)

    staging_branch = get_staging_branch_head(branchenv)
    if staging_branch == name:
        msg = f'Branch: {name} cannot be deleted while acting as the base for '\
              f'contents of the staging area. re-run after checking out a '\
              f'different branch in `write` mode.'
        raise PermissionError(msg)

    HEAD = get_branch_head_commit(branchenv, name)
    if not force_delete:
        for branch in alive_branches:
            b_head = get_branch_head_commit(branchenv, branch)
            b_ancestors = get_commit_ancestors_graph(refenv, starting_commit=b_head)
            if HEAD in b_ancestors:
                break
        else:  # N.B. for-else conditional (ie. "no break")
            msg = f'The branch {name} is not fully merged. If you are sure '\
                  f'you want to delete it, re-run with force-remove parameter set'
            raise RuntimeError(msg)

    branchtxn = TxnRegister().begin_writer_txn(branchenv)
    try:
        branchHeadKey = repo_branch_head_db_key_from_raw_key(name)
        branchtxn.delete(branchHeadKey)
    finally:
        TxnRegister().commit_writer_txn(branchenv)

    return BranchHead(name=name, digest=HEAD)


# ------------- set and get with staging area HEAD branch name --------------------------


def get_staging_branch_head(branchenv):
    """Get the name of the current staging area HEAD branch

    Parameters
    ----------
    branchenv : lmdb.Environment
        lmdb environment for the branch references

    Returns
    -------
    str
        name of the staging HEAD branch
    """
    headKey = repo_head_db_key()
    txn = TxnRegister().begin_reader_txn(branchenv)
    try:
        headBranchVal = txn.get(headKey)
    finally:
        TxnRegister().abort_reader_txn(branchenv)
    headBranch = repo_head_raw_val_from_db_val(headBranchVal)
    return headBranch


def set_staging_branch_head(branchenv, branch_name):
    """Set the writer HEAD to a branch name. Does not modify staging area contents.

    A writer-checkout must specify a branch name to use as it's ancestor. We do
    not allow a writer (or staging area) to exist in a "Detached HEAD" state. In
    order to make modifications starting from a specific commit, the user must
    create a branch with that commit hash as the specified "base".

    Parameters
    ----------
    branchenv : lmdb.Environment
        lmdb environment of the branch db.
    branch_name : str
        name of the branch to checkout.

    Returns
    -------
    bool
        if the operation was successful.

    Raises
    ------
    ValueError
        If the specified branch name does not exist.
    """
    headKey = repo_head_db_key()
    requestedHeadVal = repo_head_db_val_from_raw_val(branch_name)
    requestedBranchKey = repo_branch_head_db_key_from_raw_key(branch_name)

    branchtxn = TxnRegister().begin_writer_txn(branchenv)
    try:
        branchNameExists = branchtxn.get(requestedBranchKey, default=False)
        if branchNameExists is False:
            err = f'No branch with the name: {branch_name} exists, no-op performed'
            raise ValueError(err)
        else:
            branchtxn.put(headKey, requestedHeadVal)
            success = True
    finally:
        TxnRegister().commit_writer_txn(branchenv)

    return success


# ------------- get and set a named branch HEAD commit hash --------------------------===


def get_branch_head_commit(branchenv, branch_name):
    """Find the commit hash which corresponds to the HEAD of a particular branch.

    Parameters
    ----------
    branchenv: lmdb.Environment
        lmdb environment for the branch spec
    branch_name: str
        name of the branch to find the head commit hash for

    Returns
    -------
    str
        the commit hash of the branch head

    Raises
    ------
    ValueError
        if `branch_name` does not exist in the repository
    """
    requestedBranchKey = repo_branch_head_db_key_from_raw_key(branch_name)
    branchtxn = TxnRegister().begin_reader_txn(branchenv)
    try:
        branchNameVal = branchtxn.get(requestedBranchKey, default=False)
        if branchNameVal is False:
            err = f'branch with name: {branch_name} does not exist. cannot get head.'
            raise ValueError(err)
    finally:
        TxnRegister().abort_reader_txn(branchenv)

    commit_hash = repo_branch_head_raw_val_from_db_val(branchNameVal)
    return commit_hash


def set_branch_head_commit(branchenv, branch_name, commit_hash):
    """Update an existing branch HEAD to point to a new commit hash.

    Does not update stage or refenv contents. If the current HEAD of the branch
    == the new commit hash, no operation will occur and an exception will be
    thrown.

    Parameters
    ----------
    branchenv : lmdb.Environment
        lmdb environment where the branch records are kept
    branch_name : string
        Name of the branch to update the HEAD commit of
    commit_hash : string
        Commit hash to update the branch HEAD to point to.

    Returns
    -------
    string
        Commit hash of the new branch head if the operation was successful.

    Raises
    ------
    ValueError
        If the current HEAD is the same as the new commit hash.
    """
    currentHeadCommit = get_branch_head_commit(branchenv=branchenv, branch_name=branch_name)
    if currentHeadCommit == commit_hash:
        err = f'Current branch: {branch_name} HEAD: {currentHeadCommit} is same as the '\
              f'requested updated HEAD: {commit_hash}, no-op performed'
        raise ValueError(err)

    branchtxn = TxnRegister().begin_writer_txn(branchenv)
    try:
        branchHeadKey = repo_branch_head_db_key_from_raw_key(branch_name)
        branchHeadVal = repo_branch_head_db_val_from_raw_val(commit_hash)
        branchtxn.put(branchHeadKey, branchHeadVal)
    finally:
        TxnRegister().commit_writer_txn(branchenv)

    return commit_hash


def get_branch_names(branchenv):
    """get a list of all branches in the repository.

    Parameters
    ----------
    branchenv : lmdb.Environment
        lmdb environment storing the branch records.

    Returns
    -------
    list of str
        list of branch names active in the repository.
    """
    branchStartKey = K_BRANCH.encode()  # TODO: This is odd, why??
    branchNames = []
    branchTxn = TxnRegister().begin_reader_txn(branchenv)
    try:
        with branchTxn.cursor() as cursor:
            cursor.first()
            branchRangeExists = cursor.set_range(branchStartKey)
            while branchRangeExists:
                branchKey = cursor.key()
                if branchKey.startswith(branchStartKey):
                    name = repo_branch_head_raw_key_from_db_key(branchKey)
                    branchNames.append(name)
                    branchRangeExists = cursor.next()
                else:
                    branchRangeExists = False
        cursor.close()
    finally:
        TxnRegister().abort_reader_txn(branchenv)

    return branchNames


def commit_hash_to_branch_name_map(branchenv: lmdb.Environment) -> dict:
    """Determine branch names which map to commit hashs

    Parameters
    ----------
    branchenv : lmdb.Environment
        db where the branch references are stored

    Returns
    -------
    dict
        keys are commit hash strings, values are list of branch names (strings)
        whose HEAD are at the key commit
    """
    outMap = defaultdict(list)
    branchNames = get_branch_names(branchenv=branchenv)
    for branchName in branchNames:
        branchHEAD = get_branch_head_commit(branchenv=branchenv, branch_name=branchName)
        outMap[branchHEAD].append(branchName)

    return outMap


# ----------------------------- Remotes ---------------------------------------


def add_remote(branchenv: lmdb.Environment, name: str, address: str) -> bool:
    """add a remote server reference to the repository.

    This method does not check that the remote is actually accessible, rather it
    just records the reference. If a remote with the same name already exists,
    no change will occur.

    Parameters
    ----------
    branchenv : lmdb.Environment
        db where the branch (and remote) references are stored.
    name : str
        name of the remote to add the address for
    address : str
        IP:PORT where the remote server can be accessed

    Returns
    -------
    bool
        True if the new reference was saved, False if not.
    """
    dbKey = remote_db_key_from_raw_key(name)
    dbVal = remote_db_val_from_raw_val(address)

    branchTxn = TxnRegister().begin_writer_txn(branchenv)
    try:
        succ = branchTxn.put(dbKey, dbVal, overwrite=False)
    finally:
        TxnRegister().commit_writer_txn(branchenv)

    return succ


def get_remote_address(branchenv: lmdb.Environment, name: str) -> str:
    """Retrieve the IO:PORT of the remote server for a given name

    Parameters
    ----------
    branchenv : lmdb.Environment
        db where the branch (and remote) references are stored
    name : str
        name of the remote to fetch

    Raises
    ------
    KeyError
        if a remote with the provided name does not exist

    Returns
    -------
    str
        IP:PORT of the recorded remote server.
    """
    dbKey = remote_db_key_from_raw_key(name)
    branchTxn = TxnRegister().begin_reader_txn(branchenv)
    try:
        dbVal = branchTxn.get(dbKey, default=False)
    finally:
        TxnRegister().abort_reader_txn(branchenv)

    if dbVal is False:
        msg = f'No remote with the name: {name} exists in the repo.'
        raise KeyError(msg)
    else:
        remote_address = remote_raw_val_from_db_val(dbVal)
        return remote_address


def remove_remote(branchenv: lmdb.Environment, name: str) -> str:
    """remove a remote reference with the provided name.

    Parameters
    ----------
    branchenv : lmdb.Environment
        db where the branch (and remote) records are stored.
    name : str
        name of the remote to remove from the repo

    Raises
    ------
    ValueError
        if a remote with the provided name does not exist

    Returns
    -------
    str
        IP:PORT of the remote with provided name (which was removed)
    """
    dbKey = remote_db_key_from_raw_key(name)
    branchTxn = TxnRegister().begin_writer_txn(branchenv)
    try:
        dbVal = branchTxn.pop(dbKey)
    finally:
        TxnRegister().commit_writer_txn(branchenv)

    if dbVal is None:
        msg = f'No remote with the name: {name} exists in the repo.'
        raise ValueError(msg)

    remote_address = remote_raw_val_from_db_val(dbVal)
    return remote_address


def get_remote_names(branchenv):
    """get a list of all remotes in the repository.

    Parameters
    ----------
    branchenv : lmdb.Environment
        lmdb environment storing the branch records.

    Returns
    -------
    list of str
        list of remote names active in the repository.
    """
    remoteStartKey = K_REMOTES.encode()
    remoteNames = []
    branchTxn = TxnRegister().begin_reader_txn(branchenv)
    try:
        with branchTxn.cursor() as cursor:
            cursor.first()
            remoteRangeExists = cursor.set_range(remoteStartKey)
            while remoteRangeExists:
                remoteKey = cursor.key()
                if remoteKey.startswith(remoteStartKey):
                    name = remote_raw_key_from_db_key(remoteKey)
                    remoteNames.append(name)
                    remoteRangeExists = cursor.next()
                else:
                    remoteRangeExists = False
        cursor.close()
    finally:
        TxnRegister().abort_reader_txn(branchenv)

    return remoteNames



__version__ = '0.5.1'
__all__ = ('make_torch_dataset', 'make_tf_dataset', 'Repository')

from functools import partial
from .repository import Repository


def raise_ImportError(message, *args, **kwargs):
    raise ImportError(message)


try:
    from .dataloaders.tfloader import make_tf_dataset
except ImportError:
    make_tf_dataset = partial(raise_ImportError, "Could not import tensorflow. Install dependencies")

try:
    from .dataloaders.torchloader import make_torch_dataset
except ImportError:
    make_torch_dataset = partial(raise_ImportError, "Could not import torch. Install dependencies")




import logging
import os
import tempfile
import time
from typing import Tuple, Sequence

import blosc
import grpc
import lmdb
import numpy as np
from tqdm import tqdm

from . import chunks
from . import hangar_service_pb2
from . import hangar_service_pb2_grpc
from .header_manipulator_client_interceptor import header_adder_interceptor
from .. import constants as c
from ..context import Environments
from ..txnctx import TxnRegister
from ..backends import BACKEND_ACCESSOR_MAP, backend_decoder
from ..records import commiting
from ..records import hashs
from ..records.hashmachine import ndarray_hasher_tcode_0
from ..records import hash_data_db_key_from_raw_key
from ..records import queries
from ..records import summarize
from ..utils import set_blosc_nthreads

set_blosc_nthreads()

logger = logging.getLogger(__name__)


class HangarClient(object):
    """Client which connects and handles data transfer to the hangar server.

    Parameters
    ----------
    envs : Environments
        environment handles to manage all required calls to the local
        repostory state.
    address : str
        IP:PORT where the hangar server can be reached.
    auth_username : str, optional, kwarg-only
        credentials to use for authentication.
    auth_password : str, optional, kwarg-only, by default ''.
        credentials to use for authentication, by default ''.
    wait_for_ready : bool, optional, kwarg-only, be default True.
        If the client should wait before erring for a short period of time
        while a server is `UNAVAILABLE`, typically due to it just starting up
        at the time the connection was made
    wait_for_read_timeout : float, optional, kwarg-only, by default 5.
        If `wait_for_ready` is True, the time in seconds which the client should
        wait before raising an error. Must be positive value (greater than 0)
    """

    def __init__(self,
                 envs: Environments,
                 address: str,
                 *,
                 auth_username: str = '',
                 auth_password: str = '',
                 wait_for_ready: bool = True,
                 wait_for_ready_timeout: float = 5):

        self.env: Environments = envs
        self.address: str = address
        self.wait_ready: bool = wait_for_ready
        self.wait_ready_timeout: float = abs(wait_for_ready_timeout + 0.001)

        self.channel: grpc.Channel = None
        self.stub: hangar_service_pb2_grpc.HangarServiceStub = None
        self.header_adder_int = header_adder_interceptor(auth_username, auth_password)

        self.cfg: dict = {}
        self._rFs: BACKEND_ACCESSOR_MAP = {}

        for backend, accessor in BACKEND_ACCESSOR_MAP.items():
            if accessor is not None:
                self._rFs[backend] = accessor(
                    repo_path=self.env.repo_path,
                    schema_shape=None,
                    schema_dtype=None)
                self._rFs[backend].open(mode='r')

        self._setup_client_channel_config()

    def _setup_client_channel_config(self):
        """get grpc client configuration from server and setup channel and stub for use.
        """
        tmp_insec_channel = grpc.insecure_channel(self.address)
        tmp_channel = grpc.intercept_channel(tmp_insec_channel, self.header_adder_int)
        tmp_stub = hangar_service_pb2_grpc.HangarServiceStub(tmp_channel)
        t_init, t_tot = time.time(), 0
        while t_tot < self.wait_ready_timeout:
            try:
                request = hangar_service_pb2.GetClientConfigRequest()
                response = tmp_stub.GetClientConfig(request)
                self.cfg['push_max_nbytes'] = int(response.config['push_max_nbytes'])
                self.cfg['optimization_target'] = response.config['optimization_target']

                enable_compression = response.config['enable_compression']
                if enable_compression == 'NoCompression':
                    compression_val = grpc.Compression.NoCompression
                elif enable_compression == 'Deflate':
                    compression_val = grpc.Compression.Deflate
                elif enable_compression == 'Gzip':
                    compression_val = grpc.Compression.Gzip
                else:
                    compression_val = grpc.Compression.NoCompression
                self.cfg['enable_compression'] = compression_val

            except grpc.RpcError as err:
                if not (err.code() == grpc.StatusCode.UNAVAILABLE) and (self.wait_ready is True):
                    logger.error(err)
                    raise err
            else:
                break
            time.sleep(0.05)
            t_tot = time.time() - t_init
        else:
            err = ConnectionError(f'Server did not connect after: {self.wait_ready_timeout} sec.')
            logger.error(err)
            raise err

        tmp_channel.close()
        tmp_insec_channel.close()
        configured_channel = grpc.insecure_channel(
            self.address,
            options=[('grpc.optimization_target', self.cfg['optimization_target'])],
            compression=self.cfg['enable_compression'])
        self.channel = grpc.intercept_channel(configured_channel, self.header_adder_int)
        self.stub = hangar_service_pb2_grpc.HangarServiceStub(self.channel)

    def close(self):
        """Close reader file handles and the GRPC channel connection, invalidating this instance.
        """
        for backend_accessor in self._rFs.values():
            backend_accessor.close()
        self.channel.close()

    def ping_pong(self) -> str:
        """Ping server to ensure that connection is working

        Returns
        -------
        str
            Should be value 'PONG'
        """
        request = hangar_service_pb2.PingRequest()
        response: hangar_service_pb2.PingReply = self.stub.PING(request)
        return response.result

    def push_branch_record(self, name: str, head: str
                           ) -> hangar_service_pb2.PushBranchRecordReply:
        """Create a branch (if new) or update the server branch HEAD to new commit.

        Parameters
        ----------
        name : str
            branch name to be pushed
        head : str
            commit hash to update the server head to

        Returns
        -------
        hangar_service_pb2.PushBranchRecordReply
            code indicating success, message with human readable info
        """
        rec = hangar_service_pb2.BranchRecord(name=name, commit=head)
        request = hangar_service_pb2.PushBranchRecordRequest(rec=rec)
        response = self.stub.PushBranchRecord(request)
        return response

    def fetch_branch_record(self, name: str
                            ) -> hangar_service_pb2.FetchBranchRecordReply:
        """Get the latest head commit the server knows about for a given branch

        Parameters
        ----------
        name : str
            name of the branch to query on the server

        Returns
        -------
        hangar_service_pb2.FetchBranchRecordReply
            rec containing name and head commit if branch exists, along with
            standard error proto if it does not exist on the server.
        """
        rec = hangar_service_pb2.BranchRecord(name=name)
        request = hangar_service_pb2.FetchBranchRecordRequest(rec=rec)
        response = self.stub.FetchBranchRecord(request)
        return response

    def push_commit_record(self, commit: str, parentVal: bytes, specVal: bytes,
                           refVal: bytes
                           ) -> hangar_service_pb2.PushBranchRecordReply:
        """Push a new commit reference to the server.

        Parameters
        ----------
        commit : str
            hash digest of the commit to send
        parentVal : bytes
            lmdb ref parentVal of the commit
        specVal : bytes
            lmdb ref specVal of the commit
        refVal : bytes
            lmdb ref refVal of the commit

        Returns
        -------
        hangar_service_pb2.PushBranchRecordReply
            standard error proto
        """
        cIter = chunks.clientCommitChunkedIterator(commit=commit,
                                                   parentVal=parentVal,
                                                   specVal=specVal,
                                                   refVal=refVal)
        response = self.stub.PushCommit(cIter)
        return response

    def fetch_commit_record(self, commit: str) -> Tuple[str, bytes, bytes, bytes]:
        """get the refs for a commit digest

        Parameters
        ----------
        commit : str
            digest of the commit to retrieve the references for

        Returns
        -------
        Tuple[str, bytes, bytes, bytes]
            ['commit hash', 'parentVal', 'specVal', 'refVal']
        """
        request = hangar_service_pb2.FetchCommitRequest(commit=commit)
        replies = self.stub.FetchCommit(request)
        for idx, reply in enumerate(replies):
            if idx == 0:
                refVal = bytearray(reply.total_byte_size)
                specVal = reply.record.spec
                parentVal = reply.record.parent
                offset = 0
            size = len(reply.record.ref)
            refVal[offset: offset + size] = reply.record.ref
            offset += size

        if reply.error.code != 0:
            logger.error(reply.error)
            return False
        return (commit, parentVal, specVal, refVal)

    def fetch_schema(self, schema_hash: str) -> Tuple[str, bytes]:
        """get the schema specification for a schema hash

        Parameters
        ----------
        schema_hash : str
            schema hash to retrieve from the server

        Returns
        -------
        Tuple[str, bytes]
            ['schema hash', 'schemaVal']
        """
        schema_rec = hangar_service_pb2.SchemaRecord(digest=schema_hash)
        request = hangar_service_pb2.FetchSchemaRequest(rec=schema_rec)
        reply = self.stub.FetchSchema(request)
        if reply.error.code != 0:
            logger.error(reply.error)
            return False

        schemaVal = reply.rec.blob
        return (schema_hash, schemaVal)

    def push_schema(self, schema_hash: str,
                    schemaVal: bytes) -> hangar_service_pb2.PushSchemaReply:
        """push a schema hash record to the remote server

        Parameters
        ----------
        schema_hash : str
            hash digest of the schema being sent
        schemaVal : bytes
            ref value of the schema representation

        Returns
        -------
        hangar_service_pb2.PushSchemaReply
            standard error proto indicating success
        """
        rec = hangar_service_pb2.SchemaRecord(digest=schema_hash,
                                              blob=schemaVal)
        request = hangar_service_pb2.PushSchemaRequest(rec=rec)
        response = self.stub.PushSchema(request)
        return response

    def fetch_data(self, schema_hash: str,
                   digests: Sequence[str]) -> Sequence[Tuple[str, np.ndarray]]:
        """Fetch data hash digests for a particular schema.

        As the total size of the data to be transferred isn't known before this
        operation occurs, if more tensor data digests are requested then the
        Client is configured to allow in memory at a time, only a portion of the
        requested digests will actually be materialized. The received digests
        are listed as the return value of this function, be sure to check that
        all requested digests have been received!

        Parameters
        ----------
        schema_hash : str
            hash of the schema each of the digests is associated with
        digests : Sequence[str]
            iterable of data digests to receive

        Returns
        -------
        Sequence[Tuple[str, np.ndarray]]
            iterable containing 2-tuples' of the hash digest and np.ndarray data.

        Raises
        ------
        RuntimeError
            if received digest != requested or what was reported to be sent.
        """
        try:
            raw_digests = c.SEP_LST.join(digests).encode()
            cIter = chunks.tensorChunkedIterator(
                buf=raw_digests, uncomp_nbytes=len(raw_digests), itemsize=1,
                pb2_request=hangar_service_pb2.FetchDataRequest)

            replies = self.stub.FetchData(cIter)
            for idx, reply in enumerate(replies):
                if idx == 0:
                    uncomp_nbytes, comp_nbytes = reply.uncomp_nbytes, reply.comp_nbytes
                    dBytes, offset = bytearray(comp_nbytes), 0
                size = len(reply.raw_data)
                if size > 0:
                    dBytes[offset:offset + size] = reply.raw_data
                    offset += size
        except grpc.RpcError as rpc_error:
            if rpc_error.code() == grpc.StatusCode.RESOURCE_EXHAUSTED:
                logger.info(rpc_error.details())
            else:
                logger.error(rpc_error.details())
                raise rpc_error

        uncompBytes = blosc.decompress(dBytes)
        if uncomp_nbytes != len(uncompBytes):
            raise RuntimeError(f'uncomp_nbytes: {uncomp_nbytes} != received {comp_nbytes}')
        received_data = []
        unpacked_records = chunks.deserialize_record_pack(uncompBytes)
        for record in unpacked_records:
            data = chunks.deserialize_record(record)
            received_hash = ndarray_hasher_tcode_0(data.array)
            if received_hash != data.digest:
                raise RuntimeError(f'MANGLED! got: {received_hash} != requested: {data.digest}')
            received_data.append((received_hash, data.array))
        return received_data

    def push_data(self, schema_hash: str, digests: Sequence[str],
                  pbar: tqdm = None) -> hangar_service_pb2.PushDataReply:
        """Given a schema and digest list, read the data and send to the server

        Parameters
        ----------
        schema_hash : str
            hash of the digest schemas
        digests : Sequence[str]
            iterable of digests to be read in and sent to the server
        pbar : tqdm, optional
            progress bar instance to be updated as the operation occurs, by default None

        Returns
        -------
        hangar_service_pb2.PushDataReply
            standard error proto indicating success

        Raises
        ------
        KeyError
            if one of the input digests does not exist on the client
        rpc_error
            if the server received corrupt data
        """
        try:
            specs = []
            hashTxn = TxnRegister().begin_reader_txn(self.env.hashenv)
            for digest in digests:
                hashKey = hash_data_db_key_from_raw_key(digest)
                hashVal = hashTxn.get(hashKey, default=False)
                if not hashVal:
                    raise KeyError(f'No hash record with key: {hashKey}')
                be_loc = backend_decoder(hashVal)
                specs.append((digest, be_loc))
        finally:
            TxnRegister().abort_reader_txn(self.env.hashenv)

        try:
            totalSize, records = 0, []
            for k in self._rFs.keys():
                self._rFs[k].__enter__()
            responses = []
            for digest, spec in specs:
                arr = self._rFs[spec.backend].read_data(spec)
                record = chunks.serialize_record(arr, digest, schema_hash)
                records.append(record)
                totalSize += len(record)
                if (totalSize >= self.cfg['push_max_nbytes']) or (len(records) > 2000):
                    # send tensor pack when >= configured max nbytes occupied in memory
                    pbar.update(len(records))
                    pack = chunks.serialize_record_pack(records)
                    cIter = chunks.tensorChunkedIterator(
                        buf=pack, uncomp_nbytes=len(pack), itemsize=1,
                        pb2_request=hangar_service_pb2.PushDataRequest)
                    response = self.stub.PushData.future(cIter)
                    responses.append(response)
                    totalSize = 0
                    records = []
        except grpc.RpcError as rpc_error:
            logger.error(rpc_error.with_traceback())
            raise rpc_error
        finally:
            for k in self._rFs.keys():
                self._rFs[k].__exit__()
            if totalSize > 0:
                # finish sending all remaining tensors if max size has not been hit.
                pack = chunks.serialize_record_pack(records)
                cIter = chunks.tensorChunkedIterator(
                    buf=pack, uncomp_nbytes=len(pack), itemsize=arr.itemsize,
                    pb2_request=hangar_service_pb2.PushDataRequest)
                response = self.stub.PushData.future(cIter)
                responses.append(response)
        for fut in responses:
            last = fut.result()
        return last

    def fetch_find_missing_commits(self, branch_name):

        c_commits = commiting.list_all_commits(self.env.refenv)
        branch_rec = hangar_service_pb2.BranchRecord(name=branch_name)
        request = hangar_service_pb2.FindMissingCommitsRequest()
        request.commits.extend(c_commits)
        request.branch.CopyFrom(branch_rec)
        reply = self.stub.FetchFindMissingCommits(request)
        return reply

    def push_find_missing_commits(self, branch_name):
        branch_commits = summarize.list_history(
            refenv=self.env.refenv,
            branchenv=self.env.branchenv,
            branch_name=branch_name)
        branch_rec = hangar_service_pb2.BranchRecord(
            name=branch_name, commit=branch_commits['head'])

        request = hangar_service_pb2.FindMissingCommitsRequest()
        request.commits.extend(branch_commits['order'])
        request.branch.CopyFrom(branch_rec)
        reply = self.stub.PushFindMissingCommits(request)
        return reply

    def fetch_find_missing_hash_records(self, commit):

        all_hashs = hashs.HashQuery(self.env.hashenv).list_all_hash_keys_raw()
        all_hashs_raw = [chunks.serialize_ident(digest, '') for digest in all_hashs]
        raw_pack = chunks.serialize_record_pack(all_hashs_raw)
        pb2_func = hangar_service_pb2.FindMissingHashRecordsRequest
        cIter = chunks.missingHashRequestIterator(commit, raw_pack, pb2_func)
        responses = self.stub.FetchFindMissingHashRecords(cIter)
        for idx, response in enumerate(responses):
            if idx == 0:
                hBytes, offset = bytearray(response.total_byte_size), 0
            size = len(response.hashs)
            hBytes[offset: offset + size] = response.hashs
            offset += size

        uncompBytes = blosc.decompress(hBytes)
        raw_idents = chunks.deserialize_record_pack(uncompBytes)
        idents = [chunks.deserialize_ident(raw) for raw in raw_idents]
        return idents

    def push_find_missing_hash_records(self, commit, tmpDB: lmdb.Environment = None):

        if tmpDB is None:
            with tempfile.TemporaryDirectory() as tempD:
                tmpDF = os.path.join(tempD, 'test.lmdb')
                tmpDB = lmdb.open(path=tmpDF, **c.LMDB_SETTINGS)
                commiting.unpack_commit_ref(self.env.refenv, tmpDB, commit)
                c_hashs_schemas = queries.RecordQuery(tmpDB).data_hash_to_schema_hash()
                c_hashes = list(set(c_hashs_schemas.keys()))
                tmpDB.close()
        else:
            c_hashs_schemas = queries.RecordQuery(tmpDB).data_hash_to_schema_hash()
            c_hashes = list(set(c_hashs_schemas.keys()))

        c_hashs_raw = [chunks.serialize_ident(digest, '') for digest in c_hashes]
        raw_pack = chunks.serialize_record_pack(c_hashs_raw)
        pb2_func = hangar_service_pb2.FindMissingHashRecordsRequest
        cIter = chunks.missingHashRequestIterator(commit, raw_pack, pb2_func)

        responses = self.stub.PushFindMissingHashRecords(cIter)
        for idx, response in enumerate(responses):
            if idx == 0:
                hBytes, offset = bytearray(response.total_byte_size), 0
            size = len(response.hashs)
            hBytes[offset: offset + size] = response.hashs
            offset += size

        uncompBytes = blosc.decompress(hBytes)
        s_missing_raw = chunks.deserialize_record_pack(uncompBytes)
        s_mis_hsh = [chunks.deserialize_ident(raw).digest for raw in s_missing_raw]
        s_mis_hsh_sch = [(s_hsh, c_hashs_schemas[s_hsh]) for s_hsh in s_mis_hsh]
        return s_mis_hsh_sch

    def fetch_find_missing_schemas(self, commit):
        c_schemaset = set(hashs.HashQuery(self.env.hashenv).list_all_schema_digests())
        c_schemas = list(c_schemaset)

        request = hangar_service_pb2.FindMissingSchemasRequest()
        request.commit = commit
        request.schema_digests.extend(c_schemas)

        response = self.stub.FetchFindMissingSchemas(request)
        return response

    def push_find_missing_schemas(self, commit, tmpDB: lmdb.Environment = None):

        if tmpDB is None:
            with tempfile.TemporaryDirectory() as tempD:
                tmpDF = os.path.join(tempD, 'test.lmdb')
                tmpDB = lmdb.open(path=tmpDF, **c.LMDB_SETTINGS)
                commiting.unpack_commit_ref(self.env.refenv, tmpDB, commit)
                c_schemaset = set(queries.RecordQuery(tmpDB).schema_hashes())
                c_schemas = list(c_schemaset)
                tmpDB.close()
        else:
            c_schemaset = set(queries.RecordQuery(tmpDB).schema_hashes())
            c_schemas = list(c_schemaset)

        request = hangar_service_pb2.FindMissingSchemasRequest()
        request.commit = commit
        request.schema_digests.extend(c_schemas)

        response = self.stub.PushFindMissingSchemas(request)
        return response



from pathlib import Path
import time
from io import StringIO

import lmdb

from .commiting import (
    get_commit_ancestors_graph,
    get_commit_spec,
    tmp_cmt_env,
)
from .heads import (
    get_staging_branch_head,
    get_branch_head_commit,
    commit_hash_to_branch_name_map,
)
from .queries import RecordQuery
from .hashs import HashQuery
from ..diff import DiffOut, Changes
from ..txnctx import TxnRegister
from ..utils import format_bytes, file_size, folder_size, unique_everseen
from ..diagnostics import graphing


def log(branchenv: lmdb.Environment,
        refenv: lmdb.Environment,
        branch: str = None,
        commit: str = None,
        *,
        return_contents: bool = False,
        show_time: bool = False,
        show_user: bool = False):
    """Displays a pretty printed commit log graph to the terminal.

    .. note::

        For programatic access, the return_contents value can be set to true
        which will retrieve relevant commit specifications as dictionary
        elements.

    Parameters
    ----------
    branchenv : lmdb.Environment
        db storing information on named branch HEADS
    refenv : lmdb.Environment
        db storing full commit history refs (compressed).
    branch : str, optional
        The name of the branch to start the log process from. (Default value
        = None)
    commit : str, optional
        The commit hash to start the log process from. (Default value = None)
    return_contents : bool, optional, kwarg only
        If true, return the commit graph specifications in a dictionary
        suitable for programatic access/evaluation.
    show_time : bool, optional, kwarg only
        If true and return_contents is False, show the time of each commit
        on the printed log graph
    show_user : bool, optional, kwarg only
        If true and return_contents is False, show the committer of each
        commit on the printed log graph
    Returns
    -------
    Optional[dict]
        Dict containing the commit ancestor graph, and all specifications.
    """
    res = list_history(
        refenv=refenv,
        branchenv=branchenv,
        branch_name=branch,
        commit_hash=commit)
    branchMap = dict(commit_hash_to_branch_name_map(branchenv=branchenv))

    if return_contents:
        for digest in list(branchMap.keys()):
            if digest not in res['order']:
                del branchMap[digest]
        res['branch_heads'] = branchMap
        return res
    else:
        g = graphing.Graph()
        g.show_nodes(dag=res['ancestors'],
                     spec=res['specs'],
                     branch=branchMap,
                     start=res['head'],
                     order=res['order'],
                     show_time=show_time,
                     show_user=show_user)


def list_history(refenv, branchenv, branch_name=None, commit_hash=None):
    """Traverse commit history to specifying ancestor DAG and all ancestor specs.

    Parameters
    ----------
    refenv : lmdb.Environment
        environment containing all repository commit data.
    branchenv : lmdb.Environment
        environment containing the current staging head branch and branch head
        commit hashes
    branch_name : string, optional
        if specified, get the history starting at the head commit of this named
        branch (the default is None, which will use the `commit_hash` arg if
        available, or staging area head)
    commit_hash : string, optional
        if specified, get the history starting at this specific commit,
        overrides branch name if both are specified (the default is `None`,
        which will use the branch_name arg if available, or staging area head)

    Returns
    -------
    dict
        dict containing information about the repo history. specifies fields for
        `head`, `ancestors` (DAG of commit), and `specs` of each commit, also `order`
        encountered.
    """

    if commit_hash is not None:
        head_commit = commit_hash
    elif branch_name is not None:
        head_commit = get_branch_head_commit(branchenv=branchenv, branch_name=branch_name)
    else:
        head_branch = get_staging_branch_head(branchenv)
        head_commit = get_branch_head_commit(branchenv, head_branch)

    ancestors = get_commit_ancestors_graph(
        refenv=refenv, starting_commit=head_commit)

    commitSpecs = {}
    for commit in ancestors.keys():
        commitSpecs[commit] = dict(get_commit_spec(refenv, commit_hash=commit)._asdict())

    cmtTimeSorter = [(k, v['commit_time']) for k, v in commitSpecs.items()]
    cmtTimeSorter.sort(key=lambda t: t[1], reverse=True)
    showparentsOrder = [x[0] for x in cmtTimeSorter]

    res = {
        'head': head_commit,
        'ancestors': ancestors,
        'specs': commitSpecs,
        'order': showparentsOrder,
    }
    return res


def details(env: lmdb.Environment, line_limit=100, line_length=100) -> StringIO:  # pragma: no cover
    """Print the details of an lmdb environment to stdout

    Parameters
    ----------
    env : lmdb.Environment
        environment handle to print records of
    line_limit : int, optional
        limit to the amount of record lines printed, by default 100
    line_length : int, optional
        limit the amount of text printed per line, by default 100

    Returns
    -------
    StringIO
        buffer containing detail data.
    """
    buf = StringIO()
    buf.write('\n======================\n')
    buf.write(f'{Path(env.path()).name}\n')
    try:
        buf.write(f'File Size: {format_bytes(file_size(Path(env.path())))}\n')
    except FileNotFoundError:
        pass
    buf.write('======================\n\n')
    txn = TxnRegister().begin_reader_txn(env)
    entries = txn.stat()['entries'] - 10
    with txn.cursor() as cursor:
        count, once = 0, False
        for key, value in cursor:
            if (count >= line_limit) and (count < entries):
                count += 1
                if (once is False) and (count < entries):
                    once = True
                    buf.write('...\n...\n...\n')
                continue
            else:
                if len(value) >= line_length:
                    buf.write(f'{key} long binary\n')
                else:
                    buf.write(f'{key} {value}\n')
            count += 1
    cursor.close()
    TxnRegister().abort_reader_txn(env)
    return buf


def summary(env, *, branch='', commit='') -> StringIO:
    """Summary of data set stored in repository.

    Parameters
    ----------
    env : :class:`..context.Environments`
        class which contains all of the lmdb environments pre-initialized for use.
    commit : str
        commit hash to query. if left empty, HEAD commit is used (Default value = '')
    branch : str
        branch name to query, if left empty, HEAD will be used. (Default value = '')

    Returns
    -------
    StringIO:
        buffer formatting the contents of the commit ref at the queried commit.
    """
    if commit != '':
        cmt = commit
    elif branch != '':
        cmt = get_branch_head_commit(env.branchenv, branch)
    else:
        headBranch = get_staging_branch_head(env.branchenv)
        cmt = get_branch_head_commit(env.branchenv, headBranch)

    spec = get_commit_spec(env.refenv, cmt)._asdict()
    if cmt == '':
        buf = StringIO()
        buf.write('No commits made')
        return buf

    def _schema_digest_spec_dict(hashenv, digest):
        hq = HashQuery(hashenv)
        res = hq.get_schema_digest_spec(digest)
        return res

    with tmp_cmt_env(env.refenv, cmt) as cmtrefenv:
        query = RecordQuery(cmtrefenv)

        nbytes = folder_size(env.repo_path, recurse=True)
        humanBytes = format_bytes(nbytes)
        buf = StringIO()
        buf.write(f'Summary of Contents Contained in Data Repository \n')
        buf.write(f' \n')
        buf.write(f'================== \n')
        buf.write(f'| Repository Info \n')
        buf.write(f'|----------------- \n')
        buf.write(f'|  Base Directory: {str(env.repo_path.parent)} \n')
        buf.write(f'|  Disk Usage: {humanBytes} \n')
        buf.write(f' \n')

        buf.write(f'=================== \n')
        buf.write(f'| Commit Details \n')
        buf.write(f'------------------- \n')
        buf.write(f'|  Commit: {cmt} \n')
        buf.write(f'|  Created: {time.asctime(time.gmtime(spec["commit_time"]))} \n')
        buf.write(f'|  By: {spec["commit_user"]} \n')
        buf.write(f'|  Email: {spec["commit_email"]} \n')
        buf.write(f'|  Message: {spec["commit_message"]} \n')
        buf.write(f' \n')
        buf.write(f'================== \n')
        buf.write(f'| DataSets \n')
        buf.write(f'|----------------- \n')

        buf.write(f'|  Number of Named Columns: {query.column_count()} \n')
        for asetn, asetnSchema in query.schema_specs().items():
            buf.write(f'|\n')
            buf.write(f'|  * Column Name: {asetn} \n')
            buf.write(f'|    Num Data Pieces: {query.column_data_count(asetn.column)} \n')

            buf.write(f'|    Details: \n')
            schema_dict = _schema_digest_spec_dict(env.hashenv, asetnSchema.digest)
            for k, v in schema_dict.items():
                buf.write(f'|    - {k}: {v} \n')

    return buf


def status(hashenv: lmdb.Environment, branch_name: str, diff: DiffOut) -> StringIO:
    """Format human readable string buffer of changes in a staging area

    Parameters
    ----------
    hashenv : lmdb.Environment
        hashenv to pull usefull schema spec info from.
    branch_name : str
        Name of the branch the diff is from.
    diff : DiffOut
        diff struct tuple returned from standard diff tool.

    Returns
    -------
    StringIO
        Buffer containing human readable printable string of change summary
    """
    def _schema_digest_spec_dict(digest):
        hq = HashQuery(hashenv)
        res = hq.get_schema_digest_spec(digest)
        return res

    def _diff_info(df: Changes) -> StringIO:
        """Format buffer for each of `ADDED`, `DELETED`, `MUTATED` changes
        """
        buf = StringIO()
        buf.write(f'|---------- \n')
        buf.write(f'| Schema: {len(df.schema)} \n')
        for k, v in df.schema.items():
            digest = v.digest
            buf.write(f'|  - "{k.column}": \n')
            buf.write(f'|       digest="{digest}" \n')
            schema_spec = _schema_digest_spec_dict(digest)
            for schema_key, schema_val in schema_spec.items():
                buf.write(f'|       {schema_key}: {schema_val} \n')

        buf.write('|---------- \n')
        buf.write(f'| Samples: {len(df.samples)} \n')
        unique = unique_everseen(df.samples, lambda x: x.column)
        for u in unique:
            un = u.column
            count = sum((1 for k in df.samples if k.column == un))
            buf.write(f'|  - "{un}": {count} \n')
        buf.write(' \n')
        return buf

    buf = StringIO()
    buf.write('============ \n')
    buf.write(f'| Branch: {branch_name} \n')
    buf.write(' \n')
    for changes, changeType in zip(diff, diff.__annotations__.keys()):
        buf.write('============ \n')
        buf.write(f'| {changeType.upper()} \n')
        change_buf = _diff_info(changes)
        buf.write(change_buf.getvalue())
    return buf



"""Definition and dynamic routing to Hangar backend implementations.

This module defines the available backends for a Hangar installation & provides
dynamic routing of method calls to the appropriate backend from a stored record
specification.

Identification
--------------

A two character ascii code identifies which backend/version some record belongs
to. Valid characters are the union of ``ascii_lowercase``, ``ascii_uppercase``,
and ``ascii_digits``:

.. centered:: ``abcdefghijklmnopqrstuvwxyzABCDEFGHIJKLMNOPQRSTUVWXYZ0123456789``

Though stored as bytes in the backend, we use human readable characters (and not
unprintable bytes) to aid in human tasks like developer database dumps and
debugging. The characters making up the two digit code have the following
symantic meanings:

   *  First Character (element 0) indicates the ``backend type`` used.

   *  Second character (element 1) indicates the ``version`` of the backend type
      which should be used to parse the specification & accesss data (more on
      this later)

The number of codes possible (a 2-choice permutation with repetition) is: 3844
which we anticipate to be more then sufficient long into the future. As a
convention, the range of values in which the first digit of the code falls into
can be used to identify the storage medium location:

   *  Lowercase ``ascii_letters`` & digits ``[0, 1, 2, 3, 4]`` -> reserved for
      backends handling data on the local disk.

   *  Uppercase ``ascii_letters`` & digits ``[5, 6, 7, 8, 9]`` -> reserved for
      backends referring to data residing on a remote server.

This is not a hard and fast rule though, and can be changed in the future if the
need arises.

Process & Guarantees
--------------------

In order to maintain backwards compatibility across versions of Hangar into the
future the following ruleset is specified and MUST BE HONORED:

*  When a new backend is proposed, the contributor(s) provide the class with a
   meaningful name (``HDF5``, ``NUMPY``, ``TILEDB``, etc) identifying the
   backend to Hangar developers. The review team will provide:

   -  ``backend type`` code
   -  ``version`` code

   which all records related to that implementation identify themselves with. In
   addition, Externally facing classes / methods go by a canonical name which is
   the concatenation of the ``meaningful name`` and assigned ``"format code"``
   ie. for ``backend name: 'NUMPY'`` assigned ``type code: '1'`` and ``version
   code: '0'`` must start external method/class names with: ``NUMPY_10_foo``

*  Once a new backend is accepted, the code assigned to it is PERMANENT &
   UNCHANGING. The same code cannot be used in the future for other backends.

*  Each backend independently determines the information it needs to log/store
   to uniquely identify and retrieve a sample stored by it. There is no standard
   format, each is free to define whatever fields they find most convenient.
   Unique encode/decode methods are defined in order to serialize this
   information to bytes and then reconstruct the information later. These bytes
   are what are passed in when a retrieval request is made, and returned when a
   storage request for some piece of data is performed.

*  Once accepted, The record format specified (ie. the byte representation
   described above) cannot be modified in any way. This must remain permanent!

*  Backend (internal) methods can be updated, optimized, and/or changed at any
   time so long as:

   *  No changes to the record format specification are introduced

   *  Data stored via any previous iteration of the backend's accessor methods
      can be retrieved bitwise exactly by the "updated" version.

Before proposing a new backend or making changes to this file, please consider
reaching out to the Hangar core development team so we can guide you through the
process.
"""
import string
from typing import Dict

from .specs import (
    HDF5_00_DataHashSpec,
    HDF5_01_DataHashSpec,
    NUMPY_10_DataHashSpec,
    LMDB_30_DataHashSpec,
    REMOTE_50_DataHashSpec,
)
from .specparse import backend_decoder

from .hdf5_00 import HDF5_00_FileHandles, HDF5_00_Options
from .hdf5_01 import HDF5_01_FileHandles, HDF5_01_Options
from .lmdb_30 import LMDB_30_FileHandles, LMDB_30_Options
from .numpy_10 import NUMPY_10_FileHandles, NUMPY_10_Options
from .remote_50 import REMOTE_50_Handler, REMOTE_50_Options


BACKEND_ACCESSOR_MAP = {
    # LOCALS -> [00:50] + ['aa':'zz']
    '00': HDF5_00_FileHandles,
    '01': HDF5_01_FileHandles,
    '10': NUMPY_10_FileHandles,
    '30': LMDB_30_FileHandles,
    # REMOTES -> [50:99] + ['AA':'ZZ']
    '50': REMOTE_50_Handler,
}

BACKEND_OPTIONS_MAP = {
    '00': HDF5_00_Options,
    '01': HDF5_01_Options,
    '10': NUMPY_10_Options,
    '30': LMDB_30_Options,
    '50': REMOTE_50_Options,
}

_local_prefixes = string.digits[0:5] + string.ascii_lowercase

BACKEND_IS_LOCAL_MAP: Dict[str, bool] = {
    k: bool(k[0] in _local_prefixes) for k in BACKEND_ACCESSOR_MAP.keys()
}

__all__ = [
    'backend_decoder', 'HDF5_00_DataHashSpec', 'HDF5_01_DataHashSpec',
    'NUMPY_10_DataHashSpec', 'LMDB_30_DataHashSpec', 'REMOTE_50_DataHashSpec',
    'BACKEND_OPTIONS_MAP', 'BACKEND_ACCESSOR_MAP', 'BACKEND_IS_LOCAL_MAP',
]
